/* 
 * PVTRACE:   PV diagram envelope tracing
 *
 *	5-may-01	created			PJT
 *      6-may-01    V1.1    added clip=, center= and linear interpolation
 *                          to get the trace velocity
 *                    a     fixed RPD !!
 *     29-dec-01      b     added extra security
 *
 */


#include <stdinc.h>
#include <getparam.h>
#include <vectmath.h>
#include <filestruct.h>
#include <image.h>

string defv[] = {
  "in=???\n      Input image file - must be an PV diagram",
  "eta=0.2\n	 [0..1] weight factor for I_max adding to I_lc",
  "ilc=0\n       Lowest Contour value",
  "sign=1\n      1: Rotation largest for positive coordinates, -1: reverse",
  "sigma=0\n	 Velocity dispersion correction factor",
  "clip=\n       One (for -clip:clip) or Two clipping values",
  "vsys=0\n       System velocity",
  "inc=90\n       Inclination of disk",
  "center=0\n     Center of galaxy along position axis",
  "VERSION=1.1b\n 29-dec-01 PJT",
  NULL,
};

string usage="PV diagram envelope tracing ";

#define RPD (PI/180.0)
#ifndef HUGE
#define HUGE 1.0e20
#endif

local void pv_trace(imageptr iptr, int vsign, 
		    real eta, real ilc, real sigma, real vsys, real sini, real psys,
                    real *clip);
local int get_vels(int n, real *s, real v, real dv, real it, real *clip,
		   real *vels);


nemo_main()
{
    stream  instr;
    imageptr iptr = NULL;
    real vel, eta, ilc, sigma, sini, vsys, psys, clip[2];
    int nc, vsign = getiparam("sign");
    string mode;

    instr = stropen(getparam("in"), "r");     /* get file name and open file */
    read_image( instr, &iptr);                /* read image */
    strclose(instr);                          /* close file */

    eta = getdparam("eta");
    ilc = getdparam("ilc");
    sigma = getdparam("sigma");
    vsys = getdparam("vsys");
    sini = sin(RPD * getdparam("inc"));
    psys = getdparam("center");
    nc = nemoinpd(getparam("clip"),clip,2);
    if (nc == 1) {
      clip[1] = clip[0];
      clip[0] = -clip[0];
    } else if (nc == 0) {
      clip[0] = clip[1] = 0.0;
    } else if (nc != 2)
      error("Syntax error; need 0, 1 or 2 numbers for clip=%s",
	    getparam("clip"));

    pv_trace(iptr, vsign, eta, ilc, sigma, vsys, sini, psys, clip);
}

/*
 * TRACE:  trace around in PV diagram
 *
 */

#define MAXV  32

local void pv_trace(imageptr iptr, int vsign, 
		    real eta, real ilc, real sigma, real vsys, real sini, real psys,
		    real *clip)
{
  int  ix, iy, j, nx, ny, nv;
  real pos, v0, dv, it, imax, vel_corr;
  real *spec, *vel, vels[MAXV];

    
  nx = Nx(iptr);     /* assumed to be position for now */
  ny = Ny(iptr);     /* assumed to be velocity for now */
  spec = (real *) allocate(ny*sizeof(real));
  vel  = (real *) allocate(ny*sizeof(real));

  imax = MapMax(iptr);

  it = sqrt(sqr(eta*imax)+sqr(ilc));
  dprintf(0,"Map [%d POS x %d VEL] I_t=%g vsys=%g sini=%g\n",
	  nx,ny,it,vsys,sini);
  if (it<=0.0) error("I_t = %g too small",it);

  for (ix=0; ix<nx; ix++) {                 /* loop over all positions in P's */
    pos = ix*Dx(iptr) + Xmin(iptr) - psys;
    v0 = Ymin(iptr);
    dv = Dy(iptr);
    if (pos*vsign > 0.0) {                 /* extract spectrum that always runs 'up' */
      for (iy=0; iy<ny; iy++) {
	vel[iy] = v0 + iy*dv;
	spec[iy] = MapValue(iptr,ix,iy);
      }
    } else {
      for (iy=0; iy<ny; iy++) {
	vel[ny-iy-1] = v0 + iy*dv;
	spec[ny-iy-1] = MapValue(iptr,ix,iy);
      }
    }
    if (pos*vsign < 0.0) {           /* fix up our mini WCS header */
      v0 = v0 + (ny-1)*dv;     
      dv = -dv;
    }
    nv = get_vels(ny,spec,v0,dv,it,clip,vels);    /* get the vel's */
    printf("%g ",pos);
    vel_corr = (vels[0]-vsys)/sini;
    if (vel_corr > 0)
      vel_corr -= sigma;
    else
      vel_corr += sigma;
    printf(" %g", vel_corr);

    for (j=1; j<nv; j++) {
      printf(" %g",(vels[j]-vsys)/sini);
    }
    printf("\n");
  }
  
  free(spec);
}


/* get_vels:
 *    this spectrum is always sorted such that we're looking for
 *    a peak at the upper end of the array, the sign of 'dv' will
 *    say where this is in velocity, but the array is always running
 *    "up"
 */

local int get_vels(int n, real *s, real v0, real dv, real smin, real *clip, real *vels)
{
  real sum1=0.0, sum0=0.0, v = v0, speak = s[0], vpeak, vfit;
  real v1, v2, v3;
  int i, nret = 0, ipeak = 0, itrace = -1;

  for (i=0; i<n; i++) {       /* loop over spectrum and get basics: mom1 and peak */
    if (s[i] <= clip[0] || s[i] >= clip[1]) { /* but only outside clipping interval */
      sum0 += s[i];
      sum1 += s[i]*v;
      if (s[i] >= speak) {
	ipeak = i;
	vpeak = v;
	speak = s[i];
      }
    }
    v += dv;
  }

  for (i=n-1; i>=0; i--) {      /* find the highest pixels above smin */
    if (itrace < 0 && s[i] > smin) {
      itrace = i;
      break;
    }
  }

  /* envelope trace */
  vfit = v0 + dv*itrace;      /* pixel at which s[i] > smin */
  if (itrace > 0) {
    vfit += dv * (smin-s[itrace])/(s[itrace+1]-s[itrace]); /* small linear correction */
  } else if (itrace == 0 || itrace == n-1) {
    vfit += 0.0;    /* no correction can be made, vfit outside V range !! */
  } else
    vfit = 0.0;     /* bad itrace, should not happen */
  vels[nret++] = vfit;

  /* simple 1st order moment */
  vels[nret++] = (sum0 == 0.0 ? 0.0 : sum1/sum0);

  /* peak location */
  vels[nret++] = vpeak;

  /* peak location, from fitting paraboloid */
  if (ipeak == 0 || ipeak == n-1)
    vfit = vpeak;
  else {
    v1 = s[ipeak-1];
    v2 = s[ipeak];
    v3 = s[ipeak+1];
    if (v1+v3 == 2*v2) 
      vfit = 0.0;
    else {
      vfit = 0.5*(v1-v3)/(v1+v3-2*v2);
      vfit = dv*vfit + vpeak;
    }
  }
  vels[nret++] = vfit;

  /* AIPS++ window method */

  /* simple gaussian fit */

  return nret;
}






