// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// bcell.h                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class cell_basics                  fully inline                             |
// class basic_cell_access            fully inline                             |
// class basic_cell                   fully inline                             |
// class cell_access                  fully inline                             |
// class flagged_cell                 fully inline                             |
// some macros for looping child nodes of a cell                               |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_bcell_h
#define included_bcell_h 1

#ifndef included_flag_h
#include <nbdy/flag.h>
#endif

#ifndef included_iomanip
#include <iomanip>
#define included_iomanip
#endif

using std::ostream;
using std::setw;
using std::setprecision;

////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
  //////////////////////////////////////////////////////////////////////////////
  // nbdy::class cell_basics                                                    
  //////////////////////////////////////////////////////////////////////////////
  class cell_basics {
    cell_basics           (const cell_basics&);    // not implemented           
    cell_basics& operator=(const cell_basics&);    // not implemented           
    //--------------------------------------------------------------------------
    // friendship                                                               
    //--------------------------------------------------------------------------
    friend class basic_cell_access;                // allows data access        
    //--------------------------------------------------------------------------
    // basic types                                                              
    //--------------------------------------------------------------------------
  public:
    typedef real scalar_type;                      // type of real numbers      
    typedef vect vector_type;                      // type of vector of reals   
    //--------------------------------------------------------------------------
    // data of class cell_basics (write access through class basic_cell_access) 
    //--------------------------------------------------------------------------
  private:
    scalar_type     RADIUS;                        // half edge length of cube  
    vector_type     CENTER;                        // center of cube            
    unsigned        NUMBER;                        // # soul descendants        
    short unsigned  NSOULS;                        // # soul children           
    short unsigned  NCELLS;                        // # cell children           
    //--------------------------------------------------------------------------
    // constructor                                                              
    //--------------------------------------------------------------------------
  protected:
    cell_basics() {}
    //--------------------------------------------------------------------------
    // const data access via members                                            
    //--------------------------------------------------------------------------
    vector_type    const&center() const { return CENTER; }
    scalar_type    const&radius() const { return RADIUS; }
    unsigned       const&number() const { return NUMBER; }
    short unsigned const&nsouls() const { return NSOULS; }
    short unsigned const&ncells() const { return NCELLS; }
    //--------------------------------------------------------------------------
    // const access via friends                                                 
    //--------------------------------------------------------------------------
  public:
#define CCBCC const cell_basics*const&C
    friend bool                 has_cell_children (CCBCC){return C->NCELLS!=0; }
    friend bool                 has_soul_children (CCBCC){return C->NSOULS!=0; }
    friend bool                 is_twig           (CCBCC){return C->NCELLS==0; }
    friend vector_type    const&center            (CCBCC){return C->CENTER; }
    friend scalar_type    const&radius            (CCBCC){return C->RADIUS; }
    friend unsigned       const&number            (CCBCC){return C->NUMBER; }
    friend short unsigned const&N_cell_children   (CCBCC){return C->NCELLS; }
    friend short unsigned const&N_soul_children   (CCBCC){return C->NSOULS; }
    friend unsigned       const&N_soul_descendants(CCBCC){return C->NUMBER; }
#undef CCBCC
  };
  //////////////////////////////////////////////////////////////////////////////
  // nbdy::class basic_cell_access (needed in tree building only)               
  //////////////////////////////////////////////////////////////////////////////
  class basic_cell_access {
    //--------------------------------------------------------------------------
    // protected types and methods                                              
    //--------------------------------------------------------------------------
  protected:
    typedef cell_basics::scalar_type scalar_type;  // type of real numbers      
    typedef cell_basics::vector_type vector_type;  // type of vector of reals   
    //--------------------------------------------------------------------------
    static vector_type    &center_(cell_basics*C) { return C->CENTER; }
    static scalar_type    &radius_(cell_basics*C) { return C->RADIUS; }
    static unsigned       &number_(cell_basics*C) { return C->NUMBER; }
    static short unsigned &nsouls_(cell_basics*C) { return C->NSOULS; }
    static short unsigned &ncells_(cell_basics*C) { return C->NCELLS; }
  };
  //////////////////////////////////////////////////////////////////////////////
  // nbdy::class basic_cell<CELL,SOUL>                                          
  //////////////////////////////////////////////////////////////////////////////
  template<class CELL> class cell_access;          // forward declaration       
  template<class CELL,class SOUL>
  class basic_cell :
    public cell_basics
  {
    basic_cell           (const basic_cell&);      // not implemented           
    basic_cell& operator=(const basic_cell&);      // not implemented           
    //--------------------------------------------------------------------------
    // friendship                                                               
    //--------------------------------------------------------------------------
    friend class cell_access<CELL>;                // allows data access        
    //--------------------------------------------------------------------------
    // basic types                                                              
    //--------------------------------------------------------------------------
  public:
    typedef CELL                     cell_type;    // type of cell              
    typedef SOUL                     soul_type;    // type of soul              
    typedef typename SOUL::bodies_type bodies_type;// type of bodies            
    //--------------------------------------------------------------------------
    // data of class basic_cell (write access only through class cell_access)   
    //--------------------------------------------------------------------------
  private:
    soul_type      *CSOULS;                        // pointer to soul children  
    cell_type      *CCELLS;                        // pointer to cell children  
    //--------------------------------------------------------------------------
    // constructor                                                              
    //--------------------------------------------------------------------------
  protected:
    basic_cell() {}
    //--------------------------------------------------------------------------
    // other const and static methods                                           
    //--------------------------------------------------------------------------
  public:
    static void dump_head(ostream&o) {
      o<<" cells ncell souls nsoul number    radius            center";
    }
    //--------------------------------------------------------------------------
    void dump(ostream&o,
	      const soul_type* const &S0,
	      const cell_type* const &C0) const
    {
      if(ncells())
	o<<" "<<setw(5)<<int(static_cast<const cell_type*>(CCELLS)-C0);
      else
	o<<"     -";
      o<<" "<<setw(5)<<ncells()
       <<" "<<setw(5)<<int(static_cast<const soul_type*>(CSOULS)-S0)
       <<" "<<setw(5)<<nsouls()
       <<" "<<setw(6)<<number()
       <<" "<<setw(9)<<radius();
      for(register int i=0; i<NDIM; ++i)
	o<<" "<<setw(8)<<setprecision(4)<<center()[i];
    }
    //--------------------------------------------------------------------------
    // template<NODE> basic_cell::child<>                                       
    //--------------------------------------------------------------------------
    template<class NODE>
    class child {
    private:
      NODE* X;
    public:
      child(NODE*x)        : X(x)           {}
      child(const child&C) : X(C.X)         {}
      child&operator= (const child&C)       { X=C.X; return *this; }
      child&operator++()                    { ++X; return *this; }
      child operator++(int)                 { return child(X++); }
      child&operator+=(const int i)         { X+=i; return *this; }
      child operator+ (const int i)   const { return child(X+i); }
      child&operator--()                    { --X; return *this; }
      child operator--(int)                 { return child(X--); }
      child&operator-=(const int i)         { X-=i; return *this; }
      child operator- (const int i)   const { return child(X-i); }
      bool  operator==(const child&C) const { return X == C.X; }
      bool  operator!=(const child&C) const { return X != C.X; }
      bool  operator>=(const child&C) const { return X >= C.X; }
      bool  operator> (const child&C) const { return X >  C.X; }
      bool  operator<=(const child&C) const { return X <= C.X; }
      bool  operator< (const child&C) const { return X <  C.X; }
      bool  operator==(const NODE*S)  const { return X == S; }
      bool  operator!=(const NODE*S)  const { return X != S; }
      bool  operator>=(const NODE*S)  const { return X >= S; }
      bool  operator> (const NODE*S)  const { return X >  S; }
      bool  operator<=(const NODE*S)  const { return X <= S; }
      bool  operator< (const NODE*S)  const { return X <  S; }
      operator NODE*        ()       { return X; }
      NODE& operator*       ()       { return*X; }
      NODE* operator->      ()       { return X; }
    };
    //--------------------------------------------------------------------------
    // access to cell children                                                  
    //--------------------------------------------------------------------------
    typedef child<cell_type> cell_child;
    cell_child begin_cell_children()const{return cell_child(CCELLS);}
    cell_child end_cell_children  ()const{return cell_child(CCELLS+ncells());}
    cell_child last_cell_child    ()const{return cell_child(CCELLS+ncells()-1);}
    //--------------------------------------------------------------------------
    // access to soul children                                                  
    //--------------------------------------------------------------------------
    typedef child<soul_type> soul_child;
    soul_child begin_soul_children()const{return soul_child(CSOULS); }
    soul_child end_soul_children  ()const{return soul_child(CSOULS+nsouls()); }
    soul_child last_soul_child    ()const{return soul_child(CSOULS+nsouls()-1);}
    //--------------------------------------------------------------------------
    // access to all soul descendants                                           
    //--------------------------------------------------------------------------
    typedef child<soul_type> soul_descendant;
    soul_descendant begin_soul_descendants() const {
      return soul_descendant(CSOULS); }
    soul_descendant end_soul_descendants  () const {
      return soul_descendant(CSOULS+number()); }
    soul_descendant last_soul_descendant  () const {
      return soul_descendant(CSOULS+number()-1);}
  };
  //////////////////////////////////////////////////////////////////////////////
  // useful macros                                                              
  //////////////////////////////////////////////////////////////////////////////
  //----------------------------------------------------------------------------
  // loop the cell children of a given cell                                     
  //----------------------------------------------------------------------------
#define LoopCellKids(CELL_TYPE,                    /* type of parent cell    */\
		     CELL_PTER,                    /* pointer to parent cell */\
		     NAME)                         /* name for child cells   */\
  for(register CELL_TYPE::cell_child	           /* type of child cell     */\
      NAME  = CELL_PTER->begin_cell_children();    /* from first child       */\
      NAME != CELL_PTER->end_cell_children();      /* until beyond last      */\
    ++NAME)                                        // get next child            
  //----------------------------------------------------------------------------
  // loop the cell children of a given cell starting somewhere                  
  //----------------------------------------------------------------------------
#define LoopCellSecd(CELL_TYPE,                    /* type of parent cell    */\
		     CELL_PTER,                    /* pointer to parent cell */\
                     CELL_START,                   /* start loop here        */\
		     NAME)                         /* name for child cells   */\
  for(register CELL_TYPE::cell_child	           /* type of child cell     */\
      NAME  = CELL_START;                          /* from start child       */\
      NAME != CELL_PTER->end_cell_children();      /* until beyond last      */\
    ++NAME)                                        // get next child            
  //----------------------------------------------------------------------------
  // loop the soul children of a given cell                                     
  //----------------------------------------------------------------------------
#define LoopSoulKids(CELL_TYPE,                    /* type of parent cell    */\
		     CELL_PTER,                    /* pointer to parent cell */\
		     NAME)                         /* name for child souls   */\
  for(register CELL_TYPE::soul_child	           /* type of child cell     */\
      NAME  = CELL_PTER->begin_soul_children();    /* from first child       */\
      NAME != CELL_PTER->end_soul_children();      /* until beyond last      */\
    ++NAME)                                        // get next child            
  //----------------------------------------------------------------------------
  // loop the soul children of a given cell starting somewhere                  
  //----------------------------------------------------------------------------
#define LoopSoulSecd(CELL_TYPE,                    /* type of parent cell    */\
		     CELL_PTER,                    /* pointer to parent cell */\
                     SOUL_START,                   /* start loop here        */\
		     NAME)                         /* name for child souls   */\
  for(register CELL_TYPE::soul_child	           /* type of child cell     */\
      NAME  = SOUL_START;                          /* from start child       */\
      NAME != CELL_PTER->end_soul_children();      /* until beyond last      */\
    ++NAME)                                        // get next child            
  //----------------------------------------------------------------------------
  // loop the soul descendants of a given cell                                  
  //----------------------------------------------------------------------------
#define LoopAllSouls(CELL_TYPE,                    /* type of parent cell    */\
		     CELL_PTER,                    /* pointer to parent cell */\
		     NAME)                         /* name for child souls   */\
  for(register CELL_TYPE::soul_descendant          /* type of child cell     */\
      NAME  = CELL_PTER->begin_soul_descendants(); /* from first child       */\
      NAME != CELL_PTER->end_soul_descendants();   /* until beyond last      */\
    ++NAME)                                        // get next child            
  //----------------------------------------------------------------------------
  // loop the soul descendants of a given cell starting somewhere               
  //----------------------------------------------------------------------------
#define LoopSecSouls(CELL_TYPE,                    /* type of parent cell    */\
		     CELL_PTER,                    /* pointer to parent cell */\
                     SOUL_START,                   /* start loop here        */\
		     NAME)                         /* name for child souls   */\
  for(register CELL_TYPE::soul_descendant          /* type of child cell     */\
      NAME  = SOUL_START;                          /* from start child       */\
      NAME != CELL_PTER->end_soul_descendants();   /* until beyond last      */\
    ++NAME)                                        // get next child            
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::cell_access (needed in tree-building only)                     
  //////////////////////////////////////////////////////////////////////////////
  template<class CELL>                             // cell & soul type of tree  
  class cell_access :
    public basic_cell_access
  {
    //--------------------------------------------------------------------------
    // protected types and methods                                              
    //--------------------------------------------------------------------------
  protected:
    typedef CELL                     cell_type;    // type of associated cell   
    typedef typename CELL::soul_type soul_type;    // type of associated soul   
    //--------------------------------------------------------------------------
    static soul_type* &csouls_(basic_cell<CELL,soul_type>*C) {return C->CSOULS;}
    static cell_type* &ccells_(basic_cell<CELL,soul_type>*C) {return C->CCELLS;}
  };
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::flagged_cell<CELL,SOUL>                                        
  //////////////////////////////////////////////////////////////////////////////
  template<class CELL, class SOUL>
  class flagged_cell :
    public basic_flag,
    public basic_cell<CELL,SOUL>
  {
    flagged_cell           (const flagged_cell&);  // not implemented           
    flagged_cell& operator=(const flagged_cell&);  // not implemented           
    //--------------------------------------------------------------------------
    // constructor                                                              
    //--------------------------------------------------------------------------
  protected:
    flagged_cell() {}
    //--------------------------------------------------------------------------
    // flag manipulations                                                       
    //--------------------------------------------------------------------------
  public:
    void reset_flag() {
      basic_flag::reset_flag();
      basic_flag::add_flag(basic_flag::AL_SINK);
    }
    //--------------------------------------------------------------------------
    void reset_sink_flag() {
      basic_flag::un_flag (basic_flag::SINK);
      basic_flag::add_flag(basic_flag::AL_SINK);
    }
    //--------------------------------------------------------------------------
    void add_sink_flag(const soul_type*const&S) {
      basic_flag::add_part_flag(S,basic_flag::SINK);
      if(!is_sink(S)) basic_flag::un_flag(basic_flag::AL_SINK);
    }
    //--------------------------------------------------------------------------
    void add_sink_flag(const cell_type*const &C) {
      basic_flag::add_part_flag(C,basic_flag::SINK);
      if(!al_sink(C)) basic_flag::un_flag(basic_flag::AL_SINK);
    }
    //--------------------------------------------------------------------------
    void add_body_flag(const soul_type*const&S) {
      basic_flag::add_part_flag(S,BODY_FLAGS);
      if(!is_sink(S)) basic_flag::un_flag(AL_SINK);
    }
    //--------------------------------------------------------------------------
    void add_body_flag(const cell_type*const &C) {
      basic_flag::add_part_flag(C,BODY_FLAGS);
      if(!al_sink(C)) basic_flag::un_flag(AL_SINK);
    }
    //--------------------------------------------------------------------------
    void update_sink_flag() {
      reset_sink_flag();
      LoopCellKids(typename cell_type,this,c) add_sink_flag(c);
      LoopSoulKids(typename cell_type,this,s) add_sink_flag(s);
    }
    //--------------------------------------------------------------------------
    // other const and static methods                                           
    //--------------------------------------------------------------------------
    static void dump_head(ostream&o) {
      o<<" flg";
      basic_cell<CELL,SOUL>::dump_head(o);
    }
    //--------------------------------------------------------------------------
    void dump(ostream& o,
	      const soul_type* const &S0,
	      const cell_type* const &C0) const
    {
      o<<" "<<setw(3)<<flag_type(*this);
      basic_cell<CELL,SOUL>::dump(o,S0,C0);
    }
  };
}
////////////////////////////////////////////////////////////////////////////////
#endif // included_bcell_h
