// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// bsoul.h                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class basic_soul           fully inline                                     |
// class flagged_soul         fully inline                                     |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_bsoul_h
#define included_bsoul_h

#ifndef included_flag_h
#include <nbdy/flag.h>
#endif

#ifndef included_iomanip
#include <iomanip>
#define included_iomanip
#endif

using std::ostream;
using std::setw;
using std::setprecision;

////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::basic_soul                                                     
  //////////////////////////////////////////////////////////////////////////////
  class basic_soul {
  private:
    basic_soul           (const basic_soul&);      // not implemented           
    basic_soul& operator=(const basic_soul&);      // not implemented           
    //--------------------------------------------------------------------------
    // basic types                                                              
    //--------------------------------------------------------------------------
  public:
    typedef real        scalar_type;               // type of real numbers      
    typedef vect        vector_type;               // type of vector of reals   
    //--------------------------------------------------------------------------
    // data of class basic_cell                                                 
    //--------------------------------------------------------------------------
  private:
    vector_type COFM;                              // center of mass = position 
    unsigned    LINK;                              // associated body/index     
    //--------------------------------------------------------------------------
    // constructor                                                              
    //--------------------------------------------------------------------------
  protected:
    basic_soul() {}
    //--------------------------------------------------------------------------
    // non-const methods                                                        
    //--------------------------------------------------------------------------
    void         link(unsigned  const&L) { LINK = L; }
    vector_type &cofm()                  { return COFM; }
    scalar_type &cofm(const int i)       { return COFM[i]; }
  public:
    void copy_link (const basic_soul*S)  { LINK = S->LINK; }
    //--------------------------------------------------------------------------
    void copy_basic(const basic_soul*S)  {
      COFM = S->COFM;
      LINK = S->LINK;
    }
    //--------------------------------------------------------------------------
    // const data access via members                                            
    //--------------------------------------------------------------------------
  protected:
    vector_type const &cofm  () const { return COFM; }
    unsigned    const &mybody() const { return LINK; }
    //--------------------------------------------------------------------------
    // const data access via friends                                            
    //--------------------------------------------------------------------------
  public:
    friend vector_type const&cofm  (const basic_soul*const&S) { return S->COFM; }
    friend unsigned    const&mybody(const basic_soul*const&S) { return S->LINK; }
    //--------------------------------------------------------------------------
    // other const and static methods                                           
    //--------------------------------------------------------------------------
    static void dump_head(ostream&o) {
      o<<" mybody          position";
    }
    //--------------------------------------------------------------------------
    void dump(ostream&o) const
    {
      o<<" "<<setw(6)<<LINK;
      for(register int i=0; i!=NDIM; ++i)
	o<<" "<<setw(8)<<setprecision(4)<<cofm()[i];
    }
  };
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::flagged_soul<BODIES>                                           
  //////////////////////////////////////////////////////////////////////////////
  template<typename BODIES>
  class flagged_soul :
    public basic_flag,
    public basic_soul
  {
    flagged_soul           (const flagged_soul&);  // not implemented           
    flagged_soul& operator=(const flagged_soul&);  // not implemented           
    //--------------------------------------------------------------------------
    // basic types                                                              
    //--------------------------------------------------------------------------
  public:
    typedef BODIES bodies_type;                    // type of bodies            
    //--------------------------------------------------------------------------
    // constructor                                                              
    //--------------------------------------------------------------------------
  protected:
    flagged_soul() {}
    //--------------------------------------------------------------------------
    // flag manipulations to use with bodies                                    
    //--------------------------------------------------------------------------
  public:
    void copy_body_flag(const BODIES*B) {
      set_part_flag(&(B->flag(mybody())),basic_flag::BODY_FLAGS);
    }
    //--------------------------------------------------------------------------
    void add_body_flag(const BODIES*B) {
      add_part_flag(&(B->flag(mybody())),basic_flag::BODY_FLAGS);
    }
    //--------------------------------------------------------------------------
    void copy_sink_flag(const BODIES*B) {
      set_part_flag(&(B->flag(mybody())),basic_flag::SINK);
    }
    //--------------------------------------------------------------------------
    // flag manipulations to use with arrays                                    
    //--------------------------------------------------------------------------
    void copy_body_flag(const flag_type*F) {
      set_part_flag(F[mybody()],basic_flag::BODY_FLAGS);
    }
    //--------------------------------------------------------------------------
    void add_body_flag(const flag_type*F) {
      add_part_flag(F[mybody()],basic_flag::BODY_FLAGS);
    }
    //--------------------------------------------------------------------------
    void copy_sink_flag(const flag_type*F) {
      set_part_flag(F[mybody()],basic_flag::SINK);
    }
    //--------------------------------------------------------------------------
    // other non-const methods                                                  
    //--------------------------------------------------------------------------
    void copy(const flagged_soul*S)  {
      copy_basic(S);
      set_part_flag(S,basic_flag::BODY_FLAGS);
    }
    //--------------------------------------------------------------------------
    // const and static methods                                                 
    //--------------------------------------------------------------------------
    static void dump_head(ostream&o) {
      o<<" flg";
      basic_soul::dump_head(o);
    }
    //--------------------------------------------------------------------------
    void dump(ostream& o) const
    {
      o<<" "<<setw(3)<<flag_type(*this);
      basic_soul::dump(o);
    }
  };
}
////////////////////////////////////////////////////////////////////////////////
#endif  // included_bsoul_h
