// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// grav.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifndef included_grav_h
#define included_grav_h

#ifndef included_kern_h
#include <nbdy/kern.h>
#endif

#ifndef included_cell_h
#include <nbdy/cell.h>
#endif

#ifndef included_tayl_h
#include <nbdy/tayl.h>
#endif

#ifndef included_stat_h
#include <nbdy/stat.h>
#endif

#ifndef included_pool_h
#include <nbdy/pool.h>
#endif
////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
  //////////////////////////////////////////////////////////////////////////////
  //                                                                            
  // class nbdy::gravity_kernel                                                 
  //                                                                            
  // This class is at the heart of the algorithm. It serves as INTERACTOR in    
  // the template functions MutualInteract(), defined in iact.h, which encodes  
  // the interaction phase in a most general way.                               
  // class nbdy::gravity_kernel has member functions for soul-soul, soul-cell,  
  // cell-cell, and cell-self interactions (methods interact() below), as well  
  // as a function for the evaluation phase, method evaluate_grav().            
  //                                                                            
  // In order to allow for more efficient usage of free memory, the cells'      
  // pointer to the memory containing the Taylor coefficients is initialized    
  // only when needed, i.e. when the corresponding cell actually has non-zero   
  // Taylor coefficients. Reversely, when the memory is no longer needed, i.e.  
  // when the coefficients have been accumulated in the evaluation phase, the   
  // memory is taken back into the pool.                                        
  //                                                                            
  //////////////////////////////////////////////////////////////////////////////
  class gravity_kernel {
  private:
    //--------------------------------------------------------------------------
    // member types of class nbdy::gravity_kernel                               
    //--------------------------------------------------------------------------
    typedef cell cell_type;                            // type of cell and soul 
    typedef soul soul_type;                            // that we are dealing w.
    //--------------------------------------------------------------------------
    enum who {                                         // what subsouls are sink
      none = 0,                                        //   none                
      some = 1,                                        //   some, but not all   
      all  = 2                                         //   all                 
    };
    typedef pool<sizeof(TaylorCoeffs)> pool_type;      // pool for Taylor coeffs
    //--------------------------------------------------------------------------
    // member data of class nbdy::gravity_kernel                                
    //--------------------------------------------------------------------------
    mutable pool_type *POOL;                           // pool for cells' coeffs
    soft_type          SOFT;                           // global/individual     
    kern_type          KERN;                           // softening kernel      
    bool               CONT;                           // is kernel continuous? 
    TreeStats         *STAT;                           // statistics            
    mutable real       EPS, EPQ, IEP, IEQ;             // eps & powers of eps   
    mutable int        NC, MAXNC;
    //--------------------------------------------------------------------------
    // private member data of class nbdy::gravity_kernel                        
    //--------------------------------------------------------------------------
    real& c0              (cell* C) const { return C->coeffs()->c0(); }
    evec  c1              (cell* C) const { return C->coeffs()->c1(); }
    emat  c2              (cell* C) const { return C->coeffs()->c2(); }
    eten  c3              (cell* C) const { return C->coeffs()->c3(); }
    //--------------------------------------------------------------------------
    who who_is_sink       (const cell*) const;       
    //--------------------------------------------------------------------------
    void give_coeffs(cell*) const;
    void take_coeffs(cell*) const;
    //--------------------------------------------------------------------------
    void many_global      (const bool,soul*,const who,soul*,const soul*) const;
    void many_individual  (const bool,soul*,const who,soul*,const soul*) const;
    void many_global      (cell*, soul*) const;
    void many_individual  (cell*, soul*) const;
    void many_global      (cell*, cell*) const;
    void many_individual  (cell*, cell*) const;
    void many_global      (cell*) const;
    void many_individual  (cell*) const;
    //--------------------------------------------------------------------------
    void grav             (vect&, const real, cell*, soul*) const;
    void grav             (vect&, const real, cell*, cell*) const;
    //--------------------------------------------------------------------------
    bool mutual_global    (cell*, soul*) const;
    bool mutual_individual(cell*, soul*) const;
    bool mutual_global    (cell*, cell*) const;
    bool mutual_individual(cell*, cell*) const;
    //--------------------------------------------------------------------------
    void evaluate_grav    (cell*,TaylorGrav) const;
  public:
    //--------------------------------------------------------------------------
    // private member data of class nbdy::gravity_kernel                        
    //--------------------------------------------------------------------------
    // non-const public methods                                                 
    //--------------------------------------------------------------------------
    gravity_kernel(                                   // constructor            
		         TreeStats*t,                 // I: statistics          
		   const kern_type k,                 // I: type of kernel      
		   const uint      p=0,               // I: initial pool size   
		   const real      e=zero,            // I: softening length    
		   const soft_type s=global) :        // I: type of softening   
      POOL ( p? new pool_type(p) : 0 ),               // new pool with n coeffs 
      SOFT ( s ),                                     // set softening type     
      KERN ( k ),                                     // set softening kernel   
      CONT ( KERN==newton || KERN==p0 || KERN==p1 ||  // kernel continuous if   
	     KERN==p2 || KERN==p3 ),                  //   none of P_n          
      STAT ( t ),                                     // set statistics         
      EPS  ( e ),                                     // set softening length   
      EPQ  ( EPS*EPS ),                               // set eps^2              
      IEP  ( (EPS>0)? one/EPS : zero ),               // set 1/eps              
      IEQ  ( IEP*IEP ),                               // set 1/eps^2            
      NC   ( 0 ), MAXNC( 0 ) {}                       // set counters           
    //--------------------------------------------------------------------------
    ~gravity_kernel() {                               // destructor             
      if(POOL) delete POOL;                           //   delete pool          
    }
    //--------------------------------------------------------------------------
    void set_softening(const soft_type ST) {          // set softening type     
      SOFT=ST;                                        //   to new value         
    }
    //--------------------------------------------------------------------------
    // const public methods                                                     
    //--------------------------------------------------------------------------
    const real& current_eps  ()     const { return EPS; }
    const real& current_epsq ()     const { return EPQ; }
    bool        continuous_kernel() const { return CONT; }
    const uint& coeffs_used()       const { return MAXNC; }
    //--------------------------------------------------------------------------
    bool interact(cell*) const;                       // cell-self interaction  
    bool interact(cell*, cell*) const;                // cell-cell interaction  
    bool interact(cell*, soul*) const;                // cell-body interaction  
    void interact(soul*, soul*) const;                // body-body interaction  
    void direct_summation(cell*) const;               // direct summation       
    void evaluate_grav   (cell*) const;               // evaluation phase       
    //--------------------------------------------------------------------------
    bool split_first(const cell*A, const cell*B) const {
      return is_twig(B) || (!is_twig(A) && rmax(A) > rmax(B));
    }
    //--------------------------------------------------------------------------
    real greens_function(vect&) const;
  };
  //////////////////////////////////////////////////////////////////////////////
  // related functions                                                          
  //////////////////////////////////////////////////////////////////////////////
  inline const real& eps (const gravity_kernel* K) { return K->current_eps(); }
  inline const real& epsq(const gravity_kernel* K) { return K->current_epsq(); }
}
////////////////////////////////////////////////////////////////////////////////
#endif // included_grav_h
