// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// iact.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_iact_h
#define included_iact_h

namespace nbdy {
  //////////////////////////////////////////////////////////////////////////////
  //                                                                            
  // auxiliary classes                                                          
  //                                                                            
  //////////////////////////////////////////////////////////////////////////////
  template<class A, class B>
  struct iaction {
    A fst;                                         // first object              
    B snd;                                         // second object             
    iaction (A a=0, B b=0) : fst(a), snd(b) {}     // constructor               
    void set(A a, B b)     { fst=a;  snd=b; }      // set pair                  
  };
  //////////////////////////////////////////////////////////////////////////////
  template<class A, class B>
  class iastack {
  private:
    typedef iaction<A,B>    iact;                  // type of stack objects     
    iact    *IA, *pi;                              // first & active element    
    //--------------------------------------------------------------------------
  public:
    iastack (const unsigned int M) :               // constructor               
      IA(new iact[M]),                             //   allocate memory         
      pi(IA-1) {}                                  //   set pter to active      
    //--------------------------------------------------------------------------
    ~iastack () { delete[] IA; }                   // destructor: deallocate    
    bool is_empty() const   { return pi<IA;}       // if stack empty?           
    iact pop     ()         { return *(pi--); }    // give last active: pop     
    void push    (A a, B b) { (++pi)->set(a,b); }  // add element:      push    
  };
  //////////////////////////////////////////////////////////////////////////////
  //                                                                            
  // class nbdy::mutual_interactor<>                                            
  //                                                                            
  //////////////////////////////////////////////////////////////////////////////
  template<class INTERACTOR>
  class MutualInteractor {
  private:
    //--------------------------------------------------------------------------
    // 1 typedefs                                                               
    //--------------------------------------------------------------------------
    typedef typename INTERACTOR::cell_type CELL;     // CELL = type of cell     
    typedef typename CELL::soul_type SOUL;           // SOUL = type of soul     
    typedef iaction<CELL*,CELL*> cc_iaction;         // cell-cell interaction   
    typedef iaction<CELL*,SOUL*> cs_iaction;         // cell-soul interaction   
    typedef iastack<CELL*,CELL*> cc_stack;           // stack: cell-cell iaction
    typedef iastack<CELL*,SOUL*> cs_stack;           // stacK: cell-soul iaction
    //--------------------------------------------------------------------------
    // 2   member data of class nbdy::MutualInteractor<>                        
    //--------------------------------------------------------------------------
    const unsigned   NN;                             // size of stacks          
    mutable cc_stack CC;                             // stack for C-C iactions  
    mutable cs_stack CS;                             // stack for C-S iactions  
    const INTERACTOR*IA;                             // iactor for ind iactions 
    //--------------------------------------------------------------------------
    // 3   member methods of class nbdy::MutualInteractor<>                     
    //--------------------------------------------------------------------------
    // 3.1 private methods                                                      
    //--------------------------------------------------------------------------
#define LoopCKids(C,A)    LoopCellKids(typename CELL,C,A)
#define LoopSKids(C,A)    LoopSoulKids(typename CELL,C,A)
#define LoopCPairs(C,A,B) LoopCellSecd(typename CELL,C,A,B)
#define LoopSPairs(C,A,B) LoopSKids(C,A) LoopSoulSecd(typename CELL,C,A+1,B)
    //--------------------------------------------------------------------------
    void clear_cell_soul_stack() const {             // clear the C-S stack     
      register cs_iaction cs;                        // current C-S interaction 
      while(!CS.is_empty()) {                        // while CS non-empty >    
	cs = CS.pop();                               //   pop new C-B iaction   
	if(IA->interact(cs.fst,cs.snd)) continue;    //   success? -> next      
	LoopCKids(cs.fst,c1) CS.push(c1,cs.snd);     //   push    sub C-S       
	LoopSKids(cs.fst,s2) IA->interact(cs.snd,s2);//   perform sub S-S       
      }                                              // <                       
    }
    //--------------------------------------------------------------------------
    void clear_cell_cell_stack() const {             // clear the C-C stack     
      register cc_iaction cc;                        // current C-C interaction 
      while(!CC.is_empty()) {                        // while CC non-empty  >   
	cc = CC.pop();                               //   pop new C-C iaction   
	if(cc.fst == cc.snd) {                       //   IF self iinteraction >
	  if(IA->interact(cc.fst)) continue;         //     IF success: next    
	  LoopCKids(cc.fst,c1) {                     //     ELSE: split     >   
	    LoopSKids (cc.fst,s2)    CS.push(c1,s2); //       push sub C-S      
	    LoopCPairs(cc.fst,c1,c2) CC.push(c1,c2); //       push sub C-C      
	  }                                          //     <                   
	  LoopSPairs(cc.fst,s1,s2) IA->interact(s1,s2);//   do sub B-B          
	} else {                                     //   < ELSE mutual iaction>
	  if(IA->interact(cc.fst,cc.snd)) continue;  //     success? -> next    
	  if(IA->split_first(cc.fst,cc.snd)) {       //     IF(split 1st)     > 
	    LoopSKids(cc.fst,s2) CS.push(cc.snd,s2); //       push sub C-S      
	    LoopCKids(cc.fst,c2) CC.push(c2,cc.snd); //       push sub C-C      
	  } else {                                   //     < ELSE(split 2nd) > 
	    LoopSKids(cc.snd,s2) CS.push(cc.fst,s2); //       push sub C-S      
	    LoopCKids(cc.snd,c2) CC.push(cc.fst,c2); //       push sub C-C      
	  }                                          //     <                   
	}                                            //   <                     
	clear_cell_soul_stack();                     //   clear the CS stack    
      }                                              // <                       
    }
    //--------------------------------------------------------------------------
#undef LoopCKids
#undef LoopSKids
#undef LoopCPairs
#undef LoopSPairs
    //--------------------------------------------------------------------------
    // 3.1 public methods                                                       
    //--------------------------------------------------------------------------
  public:
    MutualInteractor(                                // constructor             
		     const INTERACTOR*iactor,        // I: interactor           
		     const unsigned   depth) :       // I: max depth            
      NN ( NSUB*(NSUB+1)*depth/2 ),                  //   set: stack size       
      CC ( NN ),                                     //   initialize cell-cell  
      CS ( NN ),                                     //   and cell-soul stacks  
      IA ( iactor ) {}                               //   set interactor        
    //--------------------------------------------------------------------------
    void cell_cell(CELL*c1, CELL*c2) const {         // cell-cell interaction   
      CC.push(c1,c2);                                // initialize stack        
      clear_cell_cell_stack();                       // work until stack empty  
    }
    //--------------------------------------------------------------------------
    void cell_self(CELL*c) const {                   // cell-self interaction   
      cell_cell(c,c);                                // just call above         
    }
    //--------------------------------------------------------------------------
    void cell_soul(CELL*c, SOUL*s) const {           // cell-soul interaction   
      CS.push(c,s);                                  // initialize stack        
      clear_cell_soul_stack();                       // work until stack empty  
    }
    //--------------------------------------------------------------------------
    void soul_soul(SOUL*s1, SOUL*s2) const {         // soul-soul interaction   
      IA->interact(s1,s2);                           // can be done directly    
    }
  };
}
////////////////////////////////////////////////////////////////////////////////
#endif // included_iact_h
