// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// pool.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class pool<>                            fully inline                        |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_pool_h
#define included_pool_h

namespace nbdy {
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::pool<>                                                         
  //                                                                            
  // - allocates blocks of K words in chunks of N elements                      
  // - single elements can be allocated and freed                               
  // - chunks are organized as linked list                                      
  // - free elements are kept in a linked list                                  
  //                                                                            
  // Do not use pool<> for K < sizeof(void*), for it becomes inefficient.       
  //////////////////////////////////////////////////////////////////////////////
  template<unsigned K>
  class pool {
  private:
    ////////////////////////////////////////////////////////////////////////////
    // 1. member types of class nbdy::pool                                      
    ////////////////////////////////////////////////////////////////////////////
    union link {
      link *NEXT;
      char  MEM[K];
    };
    struct chunk {
      link  *START;
      chunk *NEXT;
      chunk(const uint N) : START( new link[N] ), NEXT(0)
      {
	register link *l,*n;
	for(l=START, n=START+1; n!=START+N; ++l,++n) l->NEXT = n;
	l->NEXT = 0;
      }
      ~chunk() { delete START; }
    };
    ////////////////////////////////////////////////////////////////////////////
    // 2. member data of class nbdy::pool                                       
    ////////////////////////////////////////////////////////////////////////////
    uint   N;
    chunk *CHUNKS;
    link  *HEAD;
    ////////////////////////////////////////////////////////////////////////////
    // 3. member functions of class nbdy::pool                                  
    ////////////////////////////////////////////////////////////////////////////
    void grow() {
      chunk *c = new chunk(N);
      c->NEXT  = CHUNKS;
      CHUNKS   = c;
      HEAD     = CHUNKS->START;
    }
    //--------------------------------------------------------------------------
  public:
    pool(const uint n) :
      N      ( max(1u,n) ),
      CHUNKS ( new chunk(N) ),
      HEAD   ( CHUNKS->START ) {}
    //--------------------------------------------------------------------------
    ~pool() {
      register chunk *a=CHUNKS, *n;                //   actual & next chunk     
      while(a) {                                   //   WHILE(actual is valid) >
	n = a->NEXT;                               //     get next chunk        
	delete a;                                  //     delete actual chunk   
	a = n;                                     //     set actual = next     
      }                                            //   <                       
    }
    //--------------------------------------------------------------------------
    void *alloc() {
      if(HEAD==0) grow();
      link*p = HEAD;
      HEAD   = p->NEXT;
      return p;
    }
    //--------------------------------------------------------------------------
    void free(void* e) {
      link *p = static_cast<link*>(e);
      p->NEXT = HEAD;
      HEAD    = p;
    }
  };
}
////////////////////////////////////////////////////////////////////////////////
#endif // included_pool_h
