// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// Math.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 1994-1995                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_Math_h
#define included_Math_h

extern "C" {
#include <math.h>
}

#ifndef included_Pi_h
#   include <util/Pi.h>
#endif

namespace util {
  //////////////////////////////////////////////////////////////////////////////
  // (1) constants                                                              
  //////////////////////////////////////////////////////////////////////////////
  const double EulerGamma  = 0.577215664901532860606512090082;
  const double LogofTwo    = 0.693147180559945309417232121458;
  const double LogofTwoInv = 1.442695040888963407359924681002;
  const double LogofTen    = 2.302585092994045684017991454684;
  const double LogofTenInv = 0.434294481903251827651128918917;
  //////////////////////////////////////////////////////////////////////////////
  // (2) functins                                                               
  //////////////////////////////////////////////////////////////////////////////
  //============================================================================
  // miscellaneous                                                              
  //============================================================================
  void   MathError	(const char*);	    // writes out error message & exits 
  void   MathWarning	(const char*);	    // writes out warning message       
  double SphVol(const int);	            // volume: unit sphere in n dims    
  //============================================================================
  // Log's and Exp's (inlines)                                                  
  //============================================================================
  inline double ln(const double x) {
    if(x<=0) MathError("ln() of argument <= 0");
    return log(x);
  }
  //----------------------------------------------------------------------------
  inline double ld(const double x) {
    if(x<=0) MathError("ld() of argument <= 0");
    return LogofTwoInv*log(x);
  }
  //----------------------------------------------------------------------------
  inline double lg(const double x) {
    if(x<=0) MathError("lg() of argument <= 0");
    return log10(x);
  }
  //----------------------------------------------------------------------------
  inline double Tento(const double x) {
    return exp(LogofTen*x);
  }
  //----------------------------------------------------------------------------
  inline double Twoto(const double x) {
    return exp(LogofTwo*x);
  }
  //============================================================================
  // logarithms of complex trigonometric and hyperbolic functions               
  //============================================================================
#ifdef __COMPLEX__
  complex<double> lnsin (const complex<double>&);
  complex<double> lncos (const complex<double>&);
  complex<double> lnsinh(const complex<double>&);
  complex<double> lncosh(const complex<double>&);
#endif
  //============================================================================
  // Gamma functions                                                            
  //============================================================================
  // note that                                                                  
  // P(a,x) = int(exp(-t)*t^(a-1), t=0...x)                                     
  // Q(a,x) = int(exp(-t)*t^(a-1), t=x...oo)                                    
  double LogGamma(const double);                  // ln(Gamma(x))               
  double GammaP(const double, const double);      // P(a,x) :=                  
  double LogGamma(const double, const double);    // ln(Gamma(x)*Q(a,x))        
  double Loggamma(const double, const double);    // ln(Gamma(x)*P(a,x))        
#ifdef __COMPLEX__
  complex<double> LogGamma(const complex<double>);// ln(Gamma(z))               
#endif
  //============================================================================
  // Exponential integrals                                                      
  //============================================================================
  double En(const int, const double);             // E_n(x)                     
  double Ei(const double);                        // Ei(x)                      
  //============================================================================
  // Bessel functions                                                           
  //============================================================================
  double J0(const double);                        // J_0(x)                     
  double J1(const double);                        // J_1(x)                     
  double Jn(const unsigned, const double);        // J_n(x)                     
  double Y0(const double);                        // Y_0(x) [GR: N_0(x)]        
  double Y1(const double);                        // Y_1(x) [GR: N_1(x)]        
  double Yn(const unsigned, const double);        // Y_n(x) [GR: N_n(x)]        
  double I0(const double);                        // I_0(x)                     
  double I1(const double);                        // I_1(x)                     
  double In(const unsigned, const double);        // I_n(x)                     
  double K0(const double);                        // K_0(x)                     
  double K1(const double);                        // K_1(x)                     
  double Kn(const unsigned, const double);        // K_n(x)                     
  //============================================================================
  // orthogonal polynomials                                                     
  //============================================================================
  // Hermite polynomials                                                        
  //----------------------------------------------------------------------------
  double HermiteH(const unsigned, const double);       // H_n(x)                     
  void HermiteH(const unsigned, const double, double*);// H_i(x), i=0...n            
  void NormSqHermite(const unsigned, double*);         // 
  // gives the inverse squared normalization constants for the H_n(x):
  // Int dx Exp[-x^2] H_n(x) H_m(x) = N_n delta_{nm}
  double HermiteH_normalized(const unsigned, const double);
  // returns the nth Hermite polynomial
  // normalized to be orthonormal w.r.t. the weight function exp(-x^2)
  void HermiteH_normalized(const unsigned, const double, double*);
  // evaluates the Hermite polynomials 0 to n
  // normalized to be orthonormal w.r.t. the weight function exp(-x^2)
}
////////////////////////////////////////////////////////////////////////////////
#endif // included_Math_h
