// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// body.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
#include <body.h>
#include <falcON.h>
#include <nbdy/nmio.h>
#include <util/ionl.h>
#include <iomanip>

using std::cerr;
using std::setw;
using std::ostream;
using std::istream;
using namespace nbdy;
////////////////////////////////////////////////////////////////////////////////
// class bodies: NEMO I/O                                                       
////////////////////////////////////////////////////////////////////////////////
#ifdef ALLOW_NEMO
extern "C" bool within();                          // NEMO range checker        
#define streq(x,y) (strcmp((x), (y)) == 0)         // from $NEMO/inc/stdinc.h   
//------------------------------------------------------------------------------
io bodies::read_nemo_particles(                    // R: what has been read     
			       const nemo_in&input,// I: nemo input             
			       real         *time, //[I: read time]             
			       const io      want, //[I: what to read]          
			       char         *times,//[I: time range]            
			       const bool    warn) //[I: warn: missing data]    
{
  register uint i;                                 // index                     
  register real t;                                 // time                      
  register basic_body*Bi;                          // current basic_body        
  // 1. open parameter set; read N & time; close parameter set; allocate bodies 
  input.open_set(nemo_io::param);                  // open nemo parameter set   
  register uint NN = input.read_N();               //   read N                  
  if(input.is_present(nemo_io::time))              //   IF time available      >
    input.read(nemo_io::time,&t);                  //     read time             
  else if(time) {                                  //   < ELSE, but time wanted>
      nemo_io::warning("bodies::read_nemo_particles: "
		       "no time found in snapshot; defaulting to 0");
      *time = t = zero;                            //     warn & default: t=0   
  } else t = zero;                                 //   < ELSE > default: t=0  <
  input.close_set(nemo_io::param);                 // close nemo parameter set  
  if(want == io::no ||                             // IF (nothing wanted OR     
     times &&                                      //     range given AND       
     !streq (times,"all") &&                       //     not all times wanted  
     !within(t,times,0.001))                       //     AND t not in range    
    return io::no;                                 //   > return undone     <   
  if(NN != N) {                                    // IF N differs from ours >  
    N    = NN;                                     //   re-set our N            
    NSPH = 0;                                      //   re-set our N_SPH        
    if(BB)   delete[] BB;                          //   de-alloc basic_body     
    if(BS) { delete[] BS; BS = 0; }                //   re-set step support     
    if(SS) { delete[] SS; SS = 0; }                //   re-set SPH support      
    BB = new basic_body[N];                        //   allocate N basic_body   
  }                                                // <                         
  // 2. open bodies set                                                         
  input.open_set(nemo_io::bodies);                 // open nemo particle set    
  register io read = io::no;                       // what has been read        
  // 2.1 try to read body masses                                                
  if(want & io::m)                                 // IF(masses wanted)    >    
    if(input.is_present(nemo_io::mass)) {          //   IF(masses present, too)>
      input.read(nemo_io::mass);                   //     read masses           
      for(i=0,Bi=BB; i!=N; ++i,++Bi)               //     loop basic bodies   > 
	Bi->mass() = input.bodies_scl(i);          //       copy mass         < 
      read |= io::m;                               //     add mass to read      
    } else if(warn)                                //   < ELSE: warning         
      nemo_io::warning("bodies::read_nemo_particles: cannot read masses");
  // 2.2 try to read body positions and/or velocities                           
  // 2.2.1 both positions and velocitiesare are required                        
  if(want & io::xv) {                              // IF([x,v] wanted)     >    
    if(input.is_present(nemo_io::posvel)) {        //   IF([x,v] present, too) >
      input.read(nemo_io::posvel);                 //     read phases           
      for(i=0,Bi=BB; i!=N; ++i,++Bi)               //     loop basic bodies   > 
	for(register int d=0; d!=NDIM; ++d) {      //       loop dimensions   > 
	  Bi->pos()[d] = input.bodies_phs(i)[d];   //         copy x            
	  Bi->vel()[d] = input.bodies_phs(i)[NDIM+d]; //      copy v            
	}                                          //     < <                   
      read |= io::xv;                              //     add phas to read      
    } else if(input.is_present(nemo_io::pos) &&    //   < ELSE IF(positions and 
	      input.is_present(nemo_io::vel)) {    //     velocities to read)  >
      input.read(nemo_io::pos);                    //     read positions        
      input.read(nemo_io::vel);                    //     read velocities       
      for(i=0,Bi=BB; i!=N; ++i,++Bi)               //     loop basic bodies   > 
	for(register int d=0; d!=NDIM; ++d) {      //       loop dimensions   > 
	  Bi->pos()[d] = input.bodies_vec(i)[d];   //         copy x            
	  Bi->vel()[d] = input.bodies_vel(i)[d];   //         copy v            
	}                                          //     < <                   
      read |= io::xv;                              //     add phas to read      
    } else if(warn)                                //   < ELSE: warning         
      nemo_io::warning("bodies::read_nemo_particles: cannot read phases");
  // 2.2.2 only positions are required                                          
  } else if(want & io::x) {                        // < ELSE IF( x wanted)     >
    if(input.is_present(nemo_io::posvel)) {        //   IF([x,v] present, too) >
      input.read(nemo_io::posvel);                 //     read phases           
      for(i=0,Bi=BB; i!=N; ++i,++Bi)               //     loop basic bodies   > 
	for(register int d=0; d!=NDIM; ++d)        //       loop dimensions   > 
	  Bi->pos()[d] = input.bodies_vec(i)[d];   //         copy x        < < 
      read |= io::x;                               //     add poss to read      
    } else if(input.is_present(nemo_io::pos)) {    //   < ELSE IF(x present)  > 
      input.read(nemo_io::pos);                    //     read positions        
      for(i=0,Bi=BB; i!=N; ++i,++Bi)               //     loop basic bodies   > 
	for(register int d=0; d!=NDIM; ++d)        //       loop dimensions   > 
	  Bi->pos()[d] = input.bodies_vec(i)[d];   //         copy x        < < 
      read |= io::x;                               //     add poss to read      
    } else if(warn)                                //   < ELSE: warning         
      nemo_io::warning("bodies::read_nemo_particles: cannot read positions");
  // 2.2.3 only velocities are required                                         
  } else if(want & io::v) {                        // < ELSE IF( v wanted)     >
    if(input.is_present(nemo_io::posvel)) {        //   IF([x,v] present, too) >
      input.read(nemo_io::posvel);                 //     read phases           
      for(i=0,Bi=BB; i!=N; ++i,++Bi)               //     loop basic bodies   > 
	for(register int d=0; d!=NDIM; ++d)        //       loop dimensions   > 
	  Bi->vel()[d] = input.bodies_vel(i)[d];   //         copy v        < < 
      read |= io::v;                               //     add vels to read      
    } else if(input.is_present(nemo_io::vel)) {    //   < ELSE IF( v present) > 
      input.read(nemo_io::vel);                    //     read positions        
      for(i=0,Bi=BB; i!=N; ++i,++Bi)               //     loop basic bodies   > 
	for(register int d=0; d!=NDIM; ++d)        //       loop dimensions   > 
	  Bi->vel()[d] = input.bodies_vel(i)[d];   //         copy v        < < 
      read |= io::v;                               //     add vels to read      
    } else if(warn)                                //   < ELSE: warning         
      nemo_io::warning("bodies::read_nemo_particles: cannot read velocities");
  }                                                // <                         
  // 2.3 try to read accelerations                                              
  if(want & io::a)                                 // IF(acc wanted)            
    if(input.is_present(nemo_io::acc)) {           //   IF(acc present, too)    
      input.read(nemo_io::acc);                    //     read them             
      for(i=0,Bi=BB; i!=N; ++i,++Bi)               //     loop basic bodies    >
	for(register int d=0; d!=NDIM; ++d)        //       loop dimensions    >
	  acc(i)[d] = input.bodies_vec(i)[d];      //         copy accs      < <
      read |= io::a;                               //     add accs to read      
    } else if(warn)                                //   < ELSE: warning         
      nemo_io::warning("bodies::read_nemo_particles: cannot read accelerations");
  // 2.4 try to read potentials                                                 
  if(want & io::p)                                 // IF(pot wanted)            
    if(input.is_present(nemo_io::pot)){            //   IF(pot present, too)    
      input.read(nemo_io::pot);                    //     read them             
      for(i=0; i!=N; ++i)                          //     loop basic bodies     
	pot(i) = input.bodies_scl(i);              //       copy potential      
      read |= io::p;                               //     add pots to read      
    } else if(warn)                                //    < ELSE: warning        
      nemo_io::warning("bodies::read_nemo_particles: cannot read potentials");
  // 2.5 try to read densities                                                  
  if(want & io::r)                                 // IF(rho wanted)            
    if(input.is_present(nemo_io::rho)) {           //   IF(rho present, too) >  
      input.read(nemo_io::rho);                    //     read them             
      for(i=0; i!=N; ++i)                          //     loop basic bodies     
	rho(i) = input.bodies_scl(i);              //       copy potential      
      read |= io::r;                               //       add rho to read     
    } else if(warn)                                //    < ELSE: warning        
      nemo_io::warning("bodies::read_nemo_particles: cannot read densities");
  // 2.6 try to read aux data                                                   
  if(want & io::au)                                // IF(aux wanted)            
    if(input.is_present(nemo_io::aux)){            //   IF(aux present, too)    
      input.read(nemo_io::aux);                    //     read them             
      enable_aux_data();                           //     allocate memory       
      for(i=0; i!=N; ++i)                          //     loop basic bodies     
	aux(i) = input.bodies_scl(i);              //       copy potential      
      read |= io::au;                              //     add auxx to read      
    } else if(warn)                                //    < ELSE: warning        
      nemo_io::warning("bodies::read_nemo_particles: cannot read aux data");
  // 2.7 try to read eps_i                                                      
  if(want & io::e)                                 // IF(eps wanted)            
    if(input.is_present(nemo_io::eps)){            //   IF(eps present, too)    
      input.read(nemo_io::eps);                    //     read them             
      for(i=0; i!=N; ++i)                          //     loop basic bodies     
	eps(i) = input.bodies_scl(i);              //       copy potential      
      read |= io::e;                               //     add epsi to read      
    } else if(warn)                                //    < ELSE: warning        
      nemo_io::warning("bodies::read_nemo_particles: cannot read eps_i");
  // finish off                                                                 
  input.close_set(nemo_io::bodies);                // close nemo particle set   
  input.reset();                                   // reset nemo input          
  return read;                                     // return add_io: what read  
}
//------------------------------------------------------------------------------
io bodies::read_nemo_snapshot (                    // R: what has been read     
			       const nemo_in&input,// I: nemo input             
			       real         *time, //[I: read time]             
			       const io      want, //[I: what to read]          
			       char         *times,//[I: time range]            
			       const bool    warn) //[I: warn: missing data]    
{
  if(!input.is_present(nemo_io::snap))             // IF no snapshot present    
    nemo_io::error("bodies::read_nemo_snapshot: no snapshot found"); // ERROR   
  input.open_set(nemo_io::snap);                   // open nemo snapshot        
  io read = read_nemo_particles(input,time,want,   // what has been read        
				times,warn);       //   by call to above?       
  input.close_set(nemo_io::snap);                  // close nemo snapshot       
  return read;                                     // return read               
}
//------------------------------------------------------------------------------
void bodies::write_nemo_particles(nemo_out  &output,
				  const real*time,
				  const io   write) const
{
  register uint i;                                 // index                     
  register basic_body*Bi;                          // current basic_body        
  // 1. open parameter set; write out N & time; close parameter set;            
  output.open_set(nemo_io::param);                 // open a parameter set      
  output.write_N(N);                               //   write N                 
  if(time)                                         //   IF(time to be written)  
    output.write(nemo_io::time,*time);             //     write out time        
  output.close_set(nemo_io::param);                // close parameter set       
  // 2. allocate memory for output; open particle set; output data;             
  if(write == io::no) return;                      // nothing to be done        
  output.allocarrays();                            // allocate memory for x,v   
  output.allocscalar();                            // allocate memory for m...  
  output.open_set(nemo_io::bodies);                // open particles set        
  output.write(nemo_io::cart);                     // put out coord system      
  // 2.1 output masses                                                          
  if(write & io::m) {                              // IF(masses to be written) >
    for(i=0,Bi=BB; i!=N; ++i,++Bi)                 //   loop basic bodies      >
      output.bodies_scl(i) = Bi->mass();           //     copy mass into aux   <
    output.write(nemo_io::mass);                   //   write out masses        
  }                                                // <                         
  // 2.2 output positions and/or velocities                                     
  // 2.2.1 both positions and velocities to be written out                      
  if(write & io::xv) {                             // IF([x,v] to be written)  >
    for(i=0,Bi=BB; i!=N; ++i,++Bi)                 //   loop basic bodies      >
      for(register int d=0; d!=NDIM; ++ d) {       //     loop dimensions      >
	output.bodies_phs(i)[d]     =Bi->pos()[d]; //       copy pos into phases
	output.bodies_phs(i)[NDIM+d]=Bi->vel()[d]; //       copy vel into phases
      }                                            //   < <                     
    output.write(nemo_io::posvel);                 //   write out phases        
  // 2.2.2 only positions to be written out                                     
  } else if(write & io::x) {                       // <ELSE IF(x to be written)>
    for(i=0,Bi=BB; i!=N; ++i,++Bi)                 //   loop basic bodies      >
      for(register int d=0; d!=NDIM; ++ d)         //     loop dimensions      >
	output.bodies_vec(i)[d] = Bi->pos()[d];    //       copy poss        < <
    output.write(nemo_io::pos);                    //   write out poss          
  // 2.2.3 only velocities to be written out                                    
  } else if(write & io::v) {                       // <ELSE IF(v to be written)>
    for(i=0,Bi=BB; i!=N; ++i,++Bi)                 //   loop basic bodies      >
      for(register int d=0; d!=NDIM; ++ d)         //     loop dimensions      >
	output.bodies_vel(i)[d] = Bi->vel()[d];    //       copy vels        < <
    output.write(nemo_io::vel);                    //   write out vels          
  }                                                // <                         
  // 2.3 output accelerations                                                   
  if(write & io::a) {                              // IF pots for output   >    
    for(i=0,Bi=BB; i!=N; ++i,++Bi)                 //   loop basic bodies    >  
      for(register int d=0; d!=NDIM; ++ d)         //     loop dimensions      >
	output.bodies_vec(i)[d] = Bi->ACC[d];      //       copy accs        < <
    output.write(nemo_io::acc);                    //   write out accs          
  }                                                //   <                       
  // 2.4 output potential(s)                                                    
  if(write & io::p || write & io::pe) {            // IF pots for output  >     
    if(write & io::p)                              //   IF N-body pots    >     
      if(write & io::pe && PX)                     //     IF extern pots  >     
	for(i=0; i!=N; ++i)                        //       loop basic bodies   
	  output.bodies_scl(i) = pot(i) + pex(i);  //         copy potential    
      else                                         //     < ELSE >              
	for(i=0; i!=N; ++i)                        //       loop basic bodies   
	  output.bodies_scl(i) = pot(i);           //         copy potential    
    else if(write & io::pe && PX)                  //   < ELSE IF extern pots>  
      for(i=0; i!=N; ++i)                          //     loop basic bodies     
	output.bodies_scl(i) = pex(i);             //       copy potential      
    output.write(nemo_io::pot);                    //   write out potential     
  }                                                // <                         
  // 2.5 output density                                                         
  if(write & io::r) {                              // IF(rhos for output)    >  
    for(i=0,Bi=BB; i!=N; ++i,++Bi)                 //   loop basic bodies    >  
      output.bodies_scl(i) = Bi->RHO;              //     copy densities     <  
    output.write(nemo_io::rho);                    //     write out scalars     
  }                                                // <                         
  // 2.6 output aux data                                                        
  if(write & io::au) {                             // IF(auxx for output)      >
    for(i=0; i!=N; ++i)                            //   loop basic bodies      >
      output.bodies_scl(i) = aux(i);               //     copy aux data        <
    output.write(nemo_io::aux);                    //   write out scalars       
  }                                                // <                         
  // 2.7 output eps_i                                                           
  if(write & io::e) {                              // IF(eps_ifor output)      >
    for(i=0,Bi=BB; i!=N; ++i,++Bi)                 //   loop basic bodies      >
      output.bodies_scl(i) = Bi->eps();            //     copy eps_i           <
    output.write(nemo_io::eps);                    //   write out potential     
  }                                                // <                         
  // 3. close particle set & reset nemo output                                  
  output.close_set(nemo_io::bodies);               // close particles set       
  output.reset();                                  // de-allocate arrays        
}
//------------------------------------------------------------------------------
void                                               // write bodies to output    
bodies::write_nemo_snapshot (nemo_out  &output,    // I: nemo output            
			     const real*time,      //[I: write time]            
			     const io   write)     //[I: write pot/acc etc?]    
  const
{
  output.open_set(nemo_io::snap);                  // open a new nemo snapshot  
  write_nemo_particles(output,time,write);         //   write particle set      
  output.close_set(nemo_io::snap);                 // close nemo snapshot       
}
////////////////////////////////////////////////////////////////////////////////
#endif // ALLLOW_NEMO
////////////////////////////////////////////////////////////////////////////////
// class bodies: YANC I/O                                                       
////////////////////////////////////////////////////////////////////////////////
void bodies::write_yanc_ascii(ostream &out, const io write) const
{
  if(write == io::no) return;
  register uint       i;
  register basic_body*Bi;
  if(write == io::mxv) {
    out<< io::mxv <<endl;
    for(Bi=BB,i=0; Bi!=BB+N; ++Bi,++i)
      out<<Bi->mass()<<" "<<Bi->pos()<<" "<<Bi->vel()<<"\n";
  } else if(write == io::mxve) {
    out<< io::mxve <<endl;
    for(Bi=BB,i=0; Bi!=BB+N; ++Bi,++i)
      out<<Bi->mass()<<" "<<Bi->pos()<<" "<<Bi->vel()<<" "<<Bi->eps()<<"\n";
  } else {
    register io written;
    if(write & io::m)         written |= io::m;
    if(write & io::x)         written |= io::x;
    if(write & io::v)         written |= io::v;
    if(write & io::e)         written |= io::e;
    if(write & io::p)         written |= io::p;
    if(write & io::pe && PX)  written |= io::pe;
    if(write & io::a)         written |= io::a;
    if(write & io::r)         written |= io::r;
    if(write & io::au  && AU) written |= io::au;
    out<<written<<endl;
    for(Bi=BB,i=0; Bi!=BB+N; ++Bi,++i) {
      if(written & io::m)  out<<Bi->mass()<<" ";
      if(written & io::x)  out<<Bi->pos() <<" ";
      if(written & io::v)  out<<Bi->vel() <<" ";
      if(written & io::e)  out<<Bi->eps() <<" ";
      if(written & io::p)  out<<Bi->POT   <<" ";
      if(written & io::pe) out<<pex(i)    <<" ";
      if(written & io::a)  out<<Bi->ACC   <<" ";
      if(written & io::r)  out<<Bi->RHO   <<" ";
      if(written & io::au) out<<aux(i)    <<" ";
      out<<"\n";
    }
  }
  out.flush();
}
//------------------------------------------------------------------------------
io bodies::read_yanc_ascii(istream& in, const io want)
{
  register io pr;
  { register int p; in >> p; pr=p; }
  register uint       i;
  register basic_body*Bi;
  if(want == io::mxv && (pr & io::mxv)) {
    for(Bi=BB,i=0; Bi!=BB+N; ++Bi,++i) {
      in>>Bi->mass()>>Bi->pos()>>Bi->vel();
      util::SwallowRestofLine(in);
    }
    return io::mxv;
  } else if(want == io::mxve && (pr & io::mxve)) {
    for(Bi=BB,i=0; Bi!=BB+N; ++Bi,++i) {
      in>>Bi->mass()>>Bi->pos()>>Bi->vel()>>Bi->eps();
      util::SwallowRestofLine(in);
    }
    return io::mxve;
  } else {
    if(pr & io::au && want & io::au) enable_aux_data();
    if(pr & io::pe && want & io::pe) enable_external_pot();
    register io   read;
    register real rdum;
    register vect vdum;
    if(want&io::m) { if(pr&io::m) read|=io::m; else warning("cannot read mass");}
    if(want&io::x) { if(pr&io::x) read|=io::x; else warning("cannot read pos"); }
    if(want&io::v) { if(pr&io::v) read|=io::v; else warning("cannot read vel"); }
    if(want&io::e) { if(pr&io::e) read|=io::e; else warning("cannot read eps"); }
    if(want&io::e) { if(pr&io::e) read|=io::e; else warning("cannot read pot"); }
    if(want&io::pe){ if(pr&io::pe)read|=io::pe;else warning("cannot read pex"); }
    if(want&io::a) { if(pr&io::a) read|=io::a; else warning("cannot read acc"); }
    if(want&io::r) { if(pr&io::r) read|=io::r; else warning("cannot read rho"); }
    if(want&io::au){ if(pr&io::au)read|=io::au;else warning("cannot read aux"); }
    for(Bi=BB,i=0; Bi!=BB+N; ++Bi,++i) {
      if(pr & io::m) { if(want & io::m)  in>>Bi->mass(); else in>>rdum; }
      if(pr & io::x) { if(want & io::x)  in>>Bi->pos();  else in>>vdum; }
      if(pr & io::v) { if(want & io::v)  in>>Bi->vel();  else in>>vdum; }
      if(pr & io::e) { if(want & io::e)  in>>Bi->eps();  else in>>rdum; }
      if(pr & io::e) { if(want & io::e)  in>>Bi->POT;    else in>>rdum; }
      if(pr & io::pe){ if(want & io::pe) in>>pex(i);     else in>>rdum; }
      if(pr & io::a) { if(want & io::a)  in>>Bi->ACC;    else in>>vdum; }
      if(pr & io::r) { if(want & io::r)  in>>Bi->RHO;    else in>>rdum; }
      if(pr & io::au){ if(want & io::au) in>>aux(i);     else in>>rdum; }
      util::SwallowRestofLine(in);
    }
    return read;
  }
}
//------------------------------------------------------------------------------
void bodies::write_yanc_binary(ostream&out) const
{
  for(register basic_body*Bi=BB; Bi!=BB+N; ++Bi)
    out.write((char*)(&(Bi->IOD)),sizeof(iodata));
  out.flush();
}
//------------------------------------------------------------------------------
void bodies::read_yanc_binary(istream&in)
{
  for(register basic_body*Bi=BB; Bi!=BB+N; ++Bi)
    if(in) in.read((char*)(&(Bi->IOD)),sizeof(iodata));
    else return;
}
////////////////////////////////////////////////////////////////////////////////
// class bodies: other non-constant member methods                              
////////////////////////////////////////////////////////////////////////////////
void bodies::flag_all_as_sink() {
  for(register basic_body*Bi=BB; Bi!=BB+N; ++Bi)
    Bi->add_flag(basic_flag::SINK);
}
////////////////////////////////////////////////////////////////////////////////
