//-----------------------------------------------------------------------------+
//                                                                             |
//  falcON_C.cc                                                                |
//                                                                             |
//  copyright Walter Dehnen, 2000/2001                                         |
//  e-mail:   dehnen@mpia.de                                                   |
//  address:  Max-Planck Institut fr Astronomie,                              |
//            Knigstuhl 17, D69117 Heidelberg Germany                         |
//                                                                             |
//-----------------------------------------------------------------------------+

#include <falcON_C.h>
#include <falcON.h>

using namespace nbdy;

#if NDIM==3
#define POSIS   const areal*X,  const areal*Y,  const areal*Z
#define VELIS   const areal*VX, const areal*VY, const areal*VZ
#define ACCIS         areal*AX,       areal*AY,       areal*AZ
#else
#define POSIS   const areal*X,  const areal*Y
#define VELIS   const areal*VX, const areal*VY
#define ACCIS         areal*AX,       areal*AY
#endif

//==============================================================================
// auxiliary data
//==============================================================================
static falcON      *FALCON    = 0;     // pointer to `our' tree
static const areal *POS[NDIM] = {0};   // pointer to position arrays
static const areal *VEL[NDIM] = {0};   // pointer to velocity arrays
static areal       *ACC[NDIM] = {0};   // pointer to acceleration arrays
static bool         BUILT     = false; // is our tree build or not?
//==============================================================================
// auxiliary routines
//==============================================================================
inline void falcON_error(const char* name)
{
  if(FALCON==0) {
    std::cerr<<" falcON ERROR: "<<name
	     <<"() called before falcON_initialize()\n";
    exit(1);
  }
}
//------------------------------------------------------------------------------
inline int falcON_warning(const char* name)
{
  if(FALCON==0) {
    std::cerr<<" falcON WARNING: "<<name
	     <<"() called before falcON_initialize()\n";
    return 1;
  }
  return 0;
}
//------------------------------------------------------------------------------
inline void falcON_grown(const char* name)
{
  if(!BUILT) {
    std::cerr<<" faclON WARNING: "<<name<<"()"
	     <<" called before a tree has been grown\n"
	     <<"        I will grow the tree (via falcON_grow()) first\n";
    FALCON->grow();
    BUILT=true;
  }
}
//------------------------------------------------------------------------------
inline nbdy::kern_type ktype(const int K)
{
  switch(K) {
  case  0: return nbdy::f0;
  case  1: return nbdy::f1;
  case  2: return nbdy::f2;
  case  3: return nbdy::f3;
  case 10: return nbdy::k0;
  case 11: return nbdy::k1;
  case 12: return nbdy::k2;
  case 13: return nbdy::k3;
  case 20: return nbdy::p0;
  case 21: return nbdy::p1;
  case 22: return nbdy::p2;
  case 23: return nbdy::p3;
  case 99: return nbdy::newton;
  default: std::cerr<<" unknown kernel, using Newtonian greens function\n";
           return nbdy::newton;
  }
}
//------------------------------------------------------------------------------
inline int ktype(const nbdy::kern_type K)
{
  switch(K) {
  case nbdy::f0: return  0;
  case nbdy::f1: return  1;
  case nbdy::f2: return  2;
  case nbdy::f3: return  3;
  case nbdy::k0: return 10;
  case nbdy::k1: return 11;
  case nbdy::k2: return 12;
  case nbdy::k3: return 13;
  case nbdy::p0: return 20;
  case nbdy::p1: return 21;
  case nbdy::p2: return 22;
  case nbdy::p3: return 23;
  case nbdy::newton: return 99;
  default:       std::cerr<<" unknown kernel type, defaulting to newton\n";
                     return 99;
  }
}
//==============================================================================
// routines defined in treeC.h
//==============================================================================
void falcON_initialize(const int *F, const areal*M, POSIS, const areal*E, ACCIS,
		       areal*PH, areal*RH, const int N, const areal EPS,
		       const areal TH, const int K, const int DMAX)
{
  if(FALCON) delete FALCON;
  POS[0] = X; ACC[0] = AX;
  POS[1] = Y; ACC[1] = AY;
#if NDIM==3
  POS[2] = Z; ACC[2] = AZ;
#endif
  FALCON = new falcON(F,M,POS,E,ACC,PH,RH,
		      abs(N),real(EPS),real(TH),ktype(K),DMAX);
  BUILT  = false;
}
//------------------------------------------------------------------------------
void falcON_resetsoftening(const areal EPS, const int K)
{
  if(falcON_warning("falcON_resetsoftening")) return;
  FALCON->reset_softening(real(EPS),ktype(K));
}
//------------------------------------------------------------------------------
void falcON_resetopening(const areal TH)
{
  if(falcON_warning("falcON_setopening")) return;
  FALCON->reset_opening(real(TH));
}  
//------------------------------------------------------------------------------
void falcON_clearup()
{
  if(FALCON) delete FALCON;
  FALCON = 0;
  BUILT  = false;
}
//------------------------------------------------------------------------------
void falcON_grow(const int Nc)
{
  falcON_error("falcON_grow");
  FALCON->grow(Nc);
  BUILT = true;
}
//------------------------------------------------------------------------------
void falcON_grow_centered(const int Nc, const int X[NDIM])
{
  falcON_error("falcON_grow_centered");
  vect C; 
  for(register indx i=0; i<NDIM; i++) C[i] = real(X[i]);
  FALCON->grow(Nc,&C);
  BUILT = true;
}
//------------------------------------------------------------------------------
void falcON_reuse()
{
  falcON_error("falcON_reuse");
  if(!BUILT) {
    std::cerr<<" faclON WARNING: falcON_reuse()"
	     <<" called before a tree has been grown\n"
	     <<"        I will grow the tree (via falcON_grow()) instead\n";
    FALCON->grow();
    BUILT = true;
  } else
    FALCON->reuse();
}
//------------------------------------------------------------------------------
void falcON_approx_grav()
{
  falcON_error("falcON_approx_gravity");
  falcON_grown("falcON_approx_gravity");
  FALCON->approximate_gravity();
}  
//------------------------------------------------------------------------------
void falcON_estimate_rho(const int nx)
{
  falcON_error("falcON_estimate_rho");
  falcON_grown("falcON_estimate_rho");
  FALCON->estimate_rho(nx);
}  
//------------------------------------------------------------------------------
void falcON_estimate_n(const int nx)
{
  falcON_error("falcON_estimate_n");
  falcON_grown("falcON_estimate_n");
  FALCON->estimate_n(nx);
}  
//------------------------------------------------------------------------------
void falcON_iactionlist(int*I1, int*I2, const int Ni, int* Nt, const areal* S,
			const areal tau, VELIS)
{
  falcON_error("falcON_iactionlist");
  falcON_grown("falcON_iactionlist");
  nbdy::uint na;
  falcON::elem_pair *bl = new falcON::elem_pair[Ni];
  VEL[0] = VX;
  VEL[1] = VY;
#if NDIM==3
  VEL[2] = VZ;
#endif
  FALCON->make_iaction_list(bl,Ni,na,S,VEL,tau);
  const nbdy::uint iend = min(na,nbdy::uint(Ni));
  for(register nbdy::uint i=0; i<iend; i++) {
    I1[i] = bl[i][0];
    I2[i] = bl[i][1];
  }
  delete[] bl;
  *Nt=na;
}
//------------------------------------------------------------------------------
void falcON_ialist(int*I1, int*I2, const int Ni, int* Nt, const areal* S,
		   const areal tau, VELIS)
{
  falcON_error("falcON_ialist");
  falcON_grown("falcON_ialist");
  nbdy::uint na;
  falcON::elem_pair *bl = new falcON::elem_pair[Ni];
  VEL[0] = VX;
  VEL[1] = VY;
#if NDIM==3
  VEL[2] = VZ;
#endif
  FALCON->make_iaction_list(bl,Ni,na,S,VEL,tau);
  const nbdy::uint iend = min(na,nbdy::uint(Ni));
  for(register nbdy::uint i=0; i<iend; i++) {
    I1[i] = 1+bl[i][0];           // FORTRAN indices = 1 + C indices
    I2[i] = 1+bl[i][1];
  }
  delete[] bl;
  *Nt=na;
}
//------------------------------------------------------------------------------
areal falcON_root_center(const int i)
{
  if(falcON_warning("falcON_root_center")) return 0.;
  return areal((FALCON->root_center())[i]);
}
//------------------------------------------------------------------------------
areal falcON_root_radius()
{
  if(falcON_warning("falcON_root_radius")) return 0.;
  return areal((FALCON->root_radius()));
}
//------------------------------------------------------------------------------
areal falcON_current_theta(const areal m)
{
  if(falcON_warning("falcON_current_theta")) return 0.;
  return areal(FALCON->theta(real(m)));
}
//------------------------------------------------------------------------------
areal falcON_current_eps()
{
  if(falcON_warning("falcON_current_eps")) return 0.;
  return areal(FALCON->softening_length());
}
//------------------------------------------------------------------------------
int falcON_current_kernel()
{
  if(falcON_warning("falcON_current_kernel")) return 0;
  return ktype(FALCON->kernel());
}
//------------------------------------------------------------------------------
int falcON_softening()
{
  if(falcON_warning("falcON_softening")) return 0;
  switch(FALCON->softening()) {
  case nbdy::global:              return 0;
  case nbdy::individual: default: return 1;
  }
}
//------------------------------------------------------------------------------
int falcON_No_cells()
{
  if(falcON_warning("falcON_No_cells")) return 0;
  return FALCON->No_cells_used();
}
//------------------------------------------------------------------------------
int falcON_depth()
{
  if(falcON_warning("falcON_depth")) return 0;
  return FALCON->root_depth();
}
//------------------------------------------------------------------------------
void falcON_stats()
{
  if(falcON_warning("falcON_stats")) return;
  FALCON->stats(cout);
}
//------------------------------------------------------------------------------
#undef POSIS
#undef VELIS
#undef ACCIS
