// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// king.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+

#include <king.h>
#include <util/Pi.h>
#include <util/ionl.h>
#include <util/tupl.h>
#include <util/nums.h>
#include <fstream>
#include <iomanip>
extern "C" {
  double exp  (double);
  double erf  (double);
  double log10(double);
}
using namespace util;
using namespace nbdy;
//==============================================================================
// nbdy::king_model::set_up() & auxiliary data & routines                       
//==============================================================================
static double c00;                                      // used by set_up() etc 
//------------------------------------------------------------------------------
inline double rhok(const double Psi) {
  // B&T eq. 4.131 with rho_1 = 1 and sigma = 1
  const    double f1  =1.1283791670955125739, f2=2./3.; // sqrt(4/Pi)           
  if(Psi<=0.) return 0.;
  register double sPsi=sqrt(Psi);
  return exp(Psi)*erf(sPsi) - f1*sPsi*(1+f2*Psi);
}
//------------------------------------------------------------------------------
inline tupel<double,2> integrand1(const double r, const tupel<double,2> Y) {
// takes x=r as independent, Y[0]=Psi, Y[1]=Psi'
  register tupel<double,2> F;
  F[0] = Y[1];
  F[1] = (r>0.)? -FPi*rhok(Y[0])-2*Y[1]/r : -c00;
  return F;
}
//------------------------------------------------------------------------------
inline tupel<double,2> integrand2(const double Psi, const tupel<double,2> Y) {
// takes Psi as independent, Y[0]=r, Y[1]=Psi'
  register tupel<double,2> F;
  F[0] = 1./Y[1];
  F[1] = -FPi*rhok(Psi)/Y[1] -2/Y[0];
  return F;
}
//------------------------------------------------------------------------------
void king_model::setup(const unsigned n)
{
  if(N != n) {
    if(r)  delete[] r;
    if(ps) delete[] ps;
    if(m)  delete[] m;
    if(rh) delete[] rh;
    N  = n;
    r  = new double[N];
    ps = new double[N];
    m  = new double[N];
    rh = new double[N];
  }
  r [0] = 0.;
  ps[0] = Psi0;
  m [0] = 0.;
  rh[0] = rhok(Psi0);
  rh0   = rh[0];
  r0    = 1.5/sqrt(Pi*rh0);            // r_0 = sqrt(9/4PiG rho_0)
  c00   = FPi*rh0/3.;                  // 4PiG rho_0/3

  register unsigned        i,Ni=N/2;
  register double          dx=2.*r0/double(Ni-1);
  register tupel<double,2> y;
  y[0] = ps[0];
  y[1] = 0.;
  for(i=1; i<Ni; i++) {
    y     = rk4(y,r[i-1],dx,&integrand1);
    r [i] = r[i-1] + dx;
    ps[i] = y[0];
    m [i] =-y[1]*r[i]*r[i];
    rh[i] = rhok(ps[i]);
  }
  y[0] = r [i-1];
  dx   =-ps[i-1] / double(N-i);
  for(; i<N; i++) {
    y     = rk4(y,ps[i-1],dx,&integrand2);
    r [i] = y[0];
    ps[i] = ps[i-1] + dx;
    m [i] =-y[1]*r[i]*r[i];
    rh[i] = rhok(ps[i]);
  }
  c  = log10(r[N-1]/r0);
  rh[N-1] = 0.0;
  P0 =-m[N-1]/r[N-1];
}
//==============================================================================
// nbdy::king_model: reset                                                      
//==============================================================================
void king_model::reset() {
  N = 0;
  if(r)  { delete[] r;  r =0; }
  if(ps) { delete[] ps; ps=0; }
  if(m)  { delete[] m;  m =0; }
  if(rh) { delete[] rh; rh=0; }
}
//==============================================================================
// nbdy::king_model: re-scaling                                                 
//==============================================================================
void king_model::reset_scales_tidal(const double M, const double Rt) {
  rscal = Rt/r[N-1];
  Pscal = M/(m[N-1]*rscal);
  vscal = sqrt(Pscal);
  rhscl = Pscal/(rscal*rscal);
  drscl = rhscl/rscal;
  mscal = Pscal*rscal;
  sdscl = Pscal/rscal;
}
//------------------------------------------------------------------------------
void king_model::reset_scales_core(const double M, const double R0) {
  rscal = R0/r0;
  Pscal = M/(m[N-1]*rscal);
  vscal = sqrt(Pscal);
  rhscl = Pscal/(rscal*rscal);
  drscl = rhscl/rscal;
  mscal = Pscal*rscal;
  sdscl = Pscal/rscal;
}
//==============================================================================
// nbdy::king_model::rms_radius()                                               
//==============================================================================
double king_model::rms_radius() const {
  register double dM,y=0.;
  for(register unsigned i=0,j=1; j!=N; ++i,++j) {
    dM  = m[j]-m[i];
    y  += (square(r[j])+square(r[i]))*dM;
  }
  return rscal*sqrt(0.5*y/m[N-1]);
}
//==============================================================================
// nbdy::king_model::Etot()                                                     
//==============================================================================
double king_model::Etot() const {
  register double dM, psi,e=0.;
  for(register unsigned i=0,j=1; j!=N; ++i,++j) {
    dM  = m[j]-m[i];
    psi = ps[j]+ps[i];
    e  += dM*psi;
  }
  return -0.25*Pscal*mscal*(square(m[N-1])/r[N-1] + e*0.5);
}
//==============================================================================
// nbdy::king_model::write_table()                                              
//==============================================================================
using std::ios;
using std::setw;
using std::setprecision;
using std::ofstream;
//------------------------------------------------------------------------------
void king_model::write_table(const char* file) const {
  ofstream table;
  if(! open(table,file) ) return;
  table.setf(ios::left, ios::adjustfield);
  table<<"#        r     Phi(r)     rho(r)       M(r)\n";
  for(register unsigned i=0; i<N; i++)
    table<<setw(10)<<setprecision(6)<<rscal*r[i]      <<" "  // r     
	 <<setw(10)<<setprecision(6)<<Pscal*(P0-ps[i])<<" "  // Phi(r)
	 <<setw(10)<<setprecision(6)<<rhscl*rh[i]     <<" "  // rho(r)
	 <<setw(10)<<setprecision(6)<<mscal*m[i]      <<"\n";// M(r)  
  table.close();
}
//==============================================================================
// nbdy::king_model::random() & auxiliary data & methods                        
//==============================================================================
static double psi,rhi;
//------------------------------------------------------------------------------
inline double feps(const double Eps) {
  // B&T eq. 4.130 with rho_1 = 1 and sigma = 1                                 
  const double fac = 0.063493635934240969785; // (2*Pi)^(-3/2)                  
  if(Eps <= 0.) return 0.;
  return fac * (exp(-Eps)-1.);
}
//------------------------------------------------------------------------------
inline void cumdens(const double v, double& P, double& p) {
  // given v and r (via psi(r)), we return mass(<v|r)-rho and                   
  // its derivative p(v|r) = 4Pi f(v|r) v^2                                     
  const double sqh = 0.70710678118654752440,
               sq2p= 0.79788456080286535588;  // sqrt(2/Pi)                     
  register double vq=v*v;
  P = exp(psi)*erf(v*sqh)-sq2p*(vq*v/3.+v*exp(psi-0.5*vq)) - rhi;
  p = FPi * feps(psi-0.5*vq) * vq;
}
//------------------------------------------------------------------------------
double king_model::random(const double R1, const double R2,
			  double &rad, double &vel) const {
  // 1. find radius from cumulative mass                                        
  register double mi = R1*m[N-1];             // mass(<rad)                     
  rad = polev(mi,m,r, N);                     // corresponding radius           
  // 2. find velocity from cumulative mass at given radius                      
  psi = polev(mi,m,ps,N);                     // psi(rad)                       
  rhi = R2*rhok(psi);                         // mass(<vel|rad)                 
  vel = rtsafe(&cumdens,0.,sqrt(2*psi),1.e-8);
  // 3. scale them to appropriate units                                         
  rad*= rscal;
  vel*= vscal;
  // 4. return potential                                                        
  return (P0-psi)*Pscal;
}
//==============================================================================
// nbdy::king_model::SurfDens() & auxiliary data & methods                      
//==============================================================================
static double *rad, *rho, Rq, rt;
static int     K;
//------------------------------------------------------------------------------
inline double sd_integrand(const double z)
{
  register double radius = sqrt(Rq+z*z);
  return (radius >= rt)? 0.0 : polev(radius,rad,rho,K);
}
//------------------------------------------------------------------------------
double king_model::SurfDens(double R) const
{
  R    /= rscal;
  ::rt  = r[N-1];
  ::K   = N;
  if(R >= rt) return 0.;
  ::Rq  = R*R;
  ::rad = r;
  ::rho = rh;
  double S = qbulir(sd_integrand,0.0,sqrt(rt*rt-Rq),1.e-6);
  return 2*Pscal/rscal*S;
}
////////////////////////////////////////////////////////////////////////////////
