// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// YancNemo.cc                                                                 |
//                                                                             |
// copyright Walter Dehnen, 2001                                               |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie, Knigstuhl 69117 Heidelberg   |
//           Germany                                                           |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
//           YancNemo -- Yet Another N-body Code, for                          |
//                       Not Everybody Must Observe                            |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// history:                                                                    |
//                                                                             |
// v 1.0.0   28/05/2001  WD created with the help of NEMO caption PJT          |
// v 1.0.1   01/06/2001  WD bug in this file removed, added history output     |
// v 1.0.2   13/06/2001  WD Ncrit option added (before Ncrit==1)               |
// v 1.0.3   20/06/2001  WD tree::make() back to adding-leafs algorithm        |
//                          Coordinatesystem given with snapshot               |
// v 1.0.4   25/06/2001  WD bug (builder::link_tree()) removed (thanks to PJT) |
// v 1.0.5   28/06/2001  WD changes in tree::grow(). Avoid box overflow.       |
//                          bug (in builder) removed                           |
// v 1.0.6   01/08/2001  WD substantial code upgrade in the tree (falcON)      |
// v 1.0.7   02/08/2001  WD Nreuse option added (before Nreuse==0)             |
// v 1.0.8   08/08/2001  WD new data lay-out for bodies, enabling special      |
//                          treatment for leap-frog, saves memory & time (?)   |
// v 1.0.9   17/09/2001  WD added support for dynamically linked external      |
//                          potential (potential.h)                            |
// v 1.0.10  20/09.2001  WD added option ("resume") for resuming an old or     |
//                          interrupted simulation. Appends to input file      |
// v 1.0.11  21/09/2001  WD improved handling of nemo I/O; can now also read   |
//                          PosTag & VelTag instead of PhaseSpaceTag           |
// v 1.0.12  21/09/2001  WD added parameter give_acc                           |
// v 1.0.13  11/10/2001  WD changed give_pot option to allow only N-body pot   |
//                          added option startout                              |
// v 1.0.14  18/10/2001  WD interweaving interaction & evaluation phase of     |
//                          gravity approximation: saves about 20b/body        |
// v 1.0.15  23/10/2001  WD some changes in I/O handling (body, yanc)          |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef ALLOW_NEMO
#error You need "ALLOW_NEMO" to compile YancNemo
#endif

#include <yanc.h>                                     // the N-body code        
#include <nbdy/pext.h>                                // external potential     
using nbdy::io;
using nbdy::yanc;
using nbdy::nemo_pot;

#include <iostream>                                   // C++ I/O                
using std::cout;

extern "C" {
#include <nemo.h>                                     // NEMO basics            
#include <potential.h>                                // NEMO potentials        
}
//------------------------------------------------------------------------------
string defv[] = {
  "in=???\n            input file                                       ",
  "out=\n              output file;   required, unless resume=t         ",
  "tstop=10\n          integration time                                 ",
  "step=1\n            time between outputs                             ",
  "theta=0.6\n         tolerance parameter at M=M_tot                   ",
  "Nreuse=0\n          max # subsequent tree re-uses                    ",
  "Ncrit=6\n           max # bodies in un-split cells                   ",
  "eps=0.05\n          softening length                                 ",
  "kernel=21\n         softening kernel                                 ",
  "hmin=6\n            tau_min = (1/2)^hmin                             ",
  "Nlev=1\n            # time-step levels                               ",
  "Nref=16\n           # bodies in cell for density estimate            ",
  "fac=\n              factor in tau < fac / rho                        ",
  "resume=f\n          resume old simulation?  that implies:\n"
  "                   - read last snapshot from input file\n"
  "                   - append to input file (unless out given)         ",
  "give_pot=0\n        0/1/2/3: output potentials: no/N-body/extern/sum ",
  "give_acc=f\n        t/f: output accelerations                        ",
  "give_rho=f\n        t/f: output density                              ",
  "potname=\n          name of external potential                       ",
  "potpars=\n          parameters of external potential                 ",
  "potfile=\n          file required by external potential              ",
  "startout=t\n        give snapshot for t=tstart?                      ",
  "VERSION=1.0.15\n    23/October/2001 WD                               ",
  NULL};
string usage = "YancNemo -- Yet Another N-body Code, for\n"
               "            Not Everybody Must Observe    ";
//------------------------------------------------------------------------------
#ifdef DOUBLEPRECISION
# define GET_POT_REAL get_potential_double
#else
# define GET_POT_REAL get_potential_float
#endif // DOUBLEPRECISION
//------------------------------------------------------------------------------
void nemo_main()
{
  // 1. set some parameters                                                     
  register double dt_out,t_out,t_end;                 // dt_out, t_out, t_end   
  t_end  = getdparam("tstop");                        // integrate until t=t_end
  dt_out = getdparam("step");                         // output every dt_out    
  register int Nlev = getiparam("Nlev");              // # time step levels     
  if(Nlev>1 && !hasvalue("fac"))                      // IF(more than one level)
    error("fac required if Nlev > 1");                //   THEN we need fac     
  nemo_pot *pot = hasvalue("potname")?                // IF(potname given) THEN 
    new nemo_pot(GET_POT_REAL(getparam("potname"),    //   initialize external  
			      getparam("potpars"),    //   potential by potname 
			      getparam("potfile")))   //   potpars & potfile    
    : 0;                                              // ELSE: not potential    
  bool resume = getbparam("resume");                  // resume old simulation ?
  io   wr     = io::mxv;                              // add to phase output :  
  if(getiparam("give_pot") &1) wr |= io::p;           //   N-body potentials    
  if(getiparam("give_pot") &2) wr |= io::pe;          //   external potentials  
  if(getbparam("give_acc"))    wr |= io::a;           //   accelerations        
  if(getbparam("give_rho"))    wr |= io::r;           //   3D densities         
  // 2. initialize N-body integrator                                            
  yanc YANC(getparam ("in"),                          //   snapshot input       
	    true,                                     //   do nemo I/O          
	    getdparam("theta"),                       //   tolerance parameter  
	    getiparam("Nreuse"),                      //   # re-using tree      
	    getiparam("Ncrit"),                       //   min# bodies/leaf cell
	    getdparam("eps"),                         //   softening length     
	    getiparam("kernel"),                      //   softening kernel     
	    getiparam("hmin"),                        //   -log_2(time step)    
	    wr,                                       //   what to put out      
	    Nlev,                                     //   # time-step levels   
	    getiparam("Nref"),                        //   # bodies in rho-estim
	    hasvalue("fac")? getdparam("fac") : 0,    //   fac in adapting steps
	    resume,                                   //   resume old simul?    
	    pot);                                     //   external potential   
  if(!YANC.okay()) error("initialization error\n");   // Initialization okay?   
  if(t_end <  YANC.initial_time()) {                  // IF(t_end < t_start)    
    warning("tstop < t_ini: nothing to be done\n");   //   THEN we are done     
    return;                                           //   and can stop here    
  }                                                   // ENDIF                  
  // 3. initialize output streams                                               
  if(hasvalue("out"))                                 // IF(out= given) THEN    
    YANC.open_nemo(getparam("out"),resume);           //   open NEMO output     
  else if(resume)                                     // ELSE IF(resuming) THEN 
    YANC.open_nemo(0,resume);                         //   append to NEMO input 
  else                                                // ELSE(not: out, resume) 
    error("out required if resume=f");                //   error out            
  if(!YANC.nemo_is_open())                            // IF(no nemo output)     
    error("opening for snapshot output failed");      //   error out            
  YANC.describe_nemo(cout,*(ask_history()));          // put history to stdout  
  YANC.short_head   (cout);                           // header for stdout      
  YANC.short_stats  (cout);                           // statistics -> stdout   
  if(!resume && getbparam("startout"))                // IF(not resuming)       
    YANC.write_nemo();                                //   write snapshot       
  // 4. time integration                                                        
  t_out = YANC.initial_time() + 0.999999*dt_out;      // time for next output   
  while(YANC.time() < t_end) {                        // WHILE( t < t_end)      
    YANC.full_step();                                 //   make full block step 
    YANC.short_stats(cout);                           //   statistics output    
    if(YANC.time() >= t_out) {                        //   IF(t >= t_out)       
      YANC.write_nemo();                              //     write snapshot     
      t_out += dt_out;                                //     increment t_out    
    }                                                 //   ENDIF                
  }                                                   // END: WHILE             
  // 5. cleaning up (including implicit call of destructors)                    
  YANC.close_nemo();                                  // close nemo output      
  if(pot) delete pot;                                 // delete external pot    
}
//---------------------end-of-YancNemo.cc------that's-it-!----------------------
