// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// mkking.cc                                                                   |
//                                                                             |
// copyright Walter Dehnen, 2001                                               |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie, Knigstuhl 69117 Heidelberg   |
//           Germany                                                           |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// creates N-body initial conditions from a King model                         |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// history:                                                                    |
//                                                                             |
// v 1.0    11/09/2001  WD  created for nemo 3.0.4                             |
// v 1.0.1  17/09/2001  WD  changed to init_xrandom (nemo 3.0.8)               |
// v 1.0.2  19/09/2001  WD  bug fixed; centrate center of mass (option zerocm) |
// v 1.0.3  21/09/2001  WD  changes in body & nmio                             |
// v 1.0.4  26/09/2001  WD  option WD_units; more bugs fixed                   |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef ALLOW_NEMO
#error You need "ALLOW_NEMO" to compile mkking
#endif

#include <util/Pi.h>
#include <king.h>
#include <body.h>
#include <nbdy/nmio.h>

extern "C" {
#include <nemo.h>
  double xrandom(double,double);
  int    init_xrandom(string);
  double sqrt(double);
  double sin(double);
  double cos(double);
}
////////////////////////////////////////////////////////////////////////////////
string defv[] = {
  "out=???\n           output file                             ",
  "nbody=???\n         number of bodies                        ",
  "W0=???\n            W0 = Psi0/sigma                         ",
  "seed=0\n            seed for the randum number generator    ",
  "time=0\n            simulation time of snapshot             ",
  "mass=1\n            total mass                              ",
  "core_radius=1\n     core radius                             ",
  "tidal_radius=\n     tidal radius (don't use core radius)    ",
  "zerocm=t\n          centrate snapshot?                      ",
  "WD_units=f\n        input:  kpc, solar masses\n"
  "                   output: kpc, kpc/Gyr, G=1 (-> mass unit)",
  "outputs=f\n         give some global quantities to stderr   ",
  "VERSION=1.0.3\n     26/September/2001 WD                    ",
  NULL};
string usage = "mkking -- construct a King model";
////////////////////////////////////////////////////////////////////////////////
void nemo_main()
{
  //----------------------------------------------------------------------------
  // 1. create king model, i.e. integrate to get Phi(r) etc.                    
  //----------------------------------------------------------------------------
  nbdy::king_model KM(getdparam("W0"));
  //----------------------------------------------------------------------------
  // 2. rescale model to mass and core/tidal radius wanted                      
  //----------------------------------------------------------------------------
  register double mass = getdparam("mass");
  if(getbparam("WD_units")) mass /= 2.2228847e5;
  if(hasvalue("tidal_radius"))
    KM.reset_scales_tidal(mass, getdparam("tidal_radius"));
  else
    KM.reset_scales_core (mass, getdparam("core_radius"));
  if(getbparam("outputs")) {
    clog<<"# King model: "
      "M_tot = "<<KM.total_mass()  <<"\n#             "
      "r_0   = "<<KM.core_radius() <<"\n#             "
      "r_t   = "<<KM.tidal_radius()<<"\n#             "
      "r_rms = "<<KM.rms_radius()  <<"\n#             "
      "sigma = "<<KM.sigma()       <<"\n#             "
      "E_tot =" <<KM.Etot()        <<endl;
  }
  //----------------------------------------------------------------------------
  // 3. create initial conditions from King model                               
  //----------------------------------------------------------------------------
  int N = getiparam("nbody");
  nbdy::bodies BB(N);
  init_xrandom(getparam("seed"));
  const    double     m = mass/double(N);
  register double     rad,vel,cth,R,phi;
  LoopBodies(nbdy::bodies,(&BB),Bi) {
    //                                                                          
    // 3.1 set mass and get r and v                                             
    //                                                                          
    Bi.mass() = m;
    KM.random(xrandom(0.,1.), xrandom(0.,1.), rad, vel);
    //                                                                          
    // 3.2 get x,y,z                                                            
    //                                                                          
    cth       = xrandom(-1.,1.);
    R         = rad * sqrt(1.-cth*cth);
    phi       = xrandom(0.,util::TPi);
    Bi.pos(0) = R * cos(phi);
    Bi.pos(1) = R * sin(phi);
    Bi.pos(2) = rad * cth;
    //                                                                          
    // 3.3 get vx,vy,vz                                                         
    //                                                                          
    cth       = xrandom(-1.,1.);
    R         = vel * sqrt(1.-cth*cth);
    phi       = xrandom(0.,util::TPi);
    Bi.vel(0) = R * cos(phi);
    Bi.vel(1) = R * sin(phi);
    Bi.vel(2) = vel * cth;
  }
  //----------------------------------------------------------------------------
  // 4. centrate position and velocity on origin                                
  //----------------------------------------------------------------------------
  if(getbparam("zerocm")) {
    register nbdy::real iN = nbdy::one/real(N);
    register nbdy::vect cpos(nbdy::zero), cvel(nbdy::zero);
    LoopBodies(nbdy::bodies,(&BB),Bi) {
      cpos += nbdy::pos(Bi);
      cvel += nbdy::vel(Bi);
    }
    cpos *= iN;
    cvel *= iN;
    LoopBodies(nbdy::bodies,(&BB),Bi) {
      Bi.pos() -= cpos;
      Bi.vel() -= cvel;
    }
  }
  //----------------------------------------------------------------------------
  // 5. output of snapshot                                                      
  //----------------------------------------------------------------------------
  nbdy::nemo_out out(getparam("out"));
  nbdy::real time = getdparam("time");
  BB.write_nemo_snapshot(out,&time,nbdy::io::mxv);
};
